#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quantum Task
"""
import json
import os
import time
import traceback

import requests
from baidubce.auth.bce_credentials import BceCredentials
from baidubce.bce_client_configuration import BceClientConfiguration
from baidubce.services.bos.bos_client import BosClient

# Configuration of token
# This could be read from configure file or environment variable
from Quanlse import Define
from Quanlse.Define import quantumHubAddr, quantumBucket, pollInterval, sdkVersion, taskSourceQuanlse, outputPath
from Quanlse.Define import waitTaskRetrys

from Quanlse.Define.Settings import outputInfo
from Quanlse.QPlatform import Error


def invokeBackend(target, params):
    """
    Invoke the Backend Functions
    """

    try:
        ret = requests.post(
            f"{quantumHubAddr}/{target}", json=params).json()
    except Exception:
        raise Error.NetworkError(traceback.format_exc())

    if ret["error"] > 0:
        errCode = ret["error"]
        errMsg = ret["message"]
        if errCode == 401:
            raise Error.TokenError(errMsg)
        else:
            raise Error.LogicError(errMsg)

    return ret["data"]


def _retryWhileNetworkError(func):
    """
    The decorator for retrying function when network failed
    """

    def _func(*args, **kwargs):
        retryCount = 0
        while retryCount < waitTaskRetrys:
            try:
                return func(*args, **kwargs)
            except Error.NetworkError:
                # retry if that's a network related error
                # other errors will be raised
                print(f"network error for {func.__name__}, retrying, {retryCount}")
                retryCount += 1
        else:
            return func(*args, **kwargs)

    return _func


@_retryWhileNetworkError
def _getSTSToken():
    """
    Get the token to upload the file

    :return:
    """

    if not Define.hubToken:
        raise Error.ArgumentError("please provide a valid token")

    config = invokeBackend("circuit/genSTS", {"token": Define.hubToken})

    bosClient = BosClient(
        BceClientConfiguration(
            credentials=BceCredentials(
                str(
                    config['accessKeyId']),
                str(
                    config['secretAccessKey'])),
            endpoint='http://bd.bcebos.com',
            security_token=str(
                config['sessionToken'])))

    return [Define.hubToken, bosClient, config['dest']]


@_retryWhileNetworkError
def _uploadCircuit(file):
    """
    Upload the file
    """

    token, client, dest = _getSTSToken()

    client.put_object_from_file(
        quantumBucket, f'tmp/{dest}', file)

    ret = invokeBackend(
        "circuit/createCircuit",
        {
            "token": token,
            "dest": dest,
            "sdkVersion": sdkVersion,
            "source": taskSourceQuanlse
        }
    )

    return token, ret['circuitId']


def _downloadToFile(url, localFile):
    """
    Download from a url to a local file
    """

    total = 0
    with requests.get(url, stream=True) as req:
        req.raise_for_status()
        with open(localFile, 'wb') as fObj:
            for chunk in req.iter_content(chunk_size=8192):
                if chunk:  # filter out keep-alive new chunks
                    fObj.write(chunk)
                    total += len(chunk)
                    # f.flush()
    return os.path.realpath(localFile), total


@_retryWhileNetworkError
def _fetchResult(token, taskId):
    """
    Fetch the result files from the taskId
    """

    params = {"token": token, "taskId": taskId}

    ret = invokeBackend("task/getTaskInfo", params)

    result = ret["result"]

    originUrl = result["originUrl"]
    # originSize = result["originSize"]
    try:
        originFile, downSize = _downloadToFile(originUrl, os.path.join(outputPath, f"remote.{taskId}.origin.json"))
    except Exception:
        # TODO split the disk write error
        raise Error.NetworkError(traceback.format_exc())
    if outputInfo:
        print(f"Result downloaded successfully {originFile} size = {downSize}")

    return originFile


def _fetchMeasureResult(taskId):
    """
    Dump the measurement content of the file from taskId
    """

    localFile = os.path.join(outputPath, f'remote.{taskId}.origin.json')
    if os.path.exists(localFile):
        with open(localFile, "rb") as fObj:
            data = json.loads(fObj.read())
            return data
    else:
        return None


@_retryWhileNetworkError
def _waitTask(token, taskId, downloadResult=True):
    """
    Wait for a task from the taskId
    """
    if outputInfo:
        print(f'Task {taskId} is running, please wait...')

    task = {
        "token": token,
        "taskId": taskId
    }

    stepStatus = "waiting"
    while True:
        try:
            time.sleep(pollInterval)
            ret = invokeBackend("task/checkTask", task)
            if ret["status"] in ("success", "failed", "manual_term"):
                if outputInfo:
                    print(f"status changed {stepStatus} => {ret['status']}")
                stepStatus = ret["status"]
                result = {"status": ret["status"]}

                if ret["status"] == "success" and "originUrl" in ret.get("result", {}):
                    if downloadResult:
                        originFile = _fetchResult(token, taskId)
                        result["origin"] = originFile
                        result["results"] = _fetchMeasureResult(taskId)
                    break
                elif ret["status"] == "failed":
                    result = ret["reason"]
                    break
                elif ret["status"] == "manual_term":
                    break
                else:
                    # go on loop
                    pass
            else:
                if ret["status"] == stepStatus:
                    continue

                if outputInfo:
                    print(f"status changed {stepStatus} => {ret['status']}")
                stepStatus = ret["status"]

        except Error.Error as err:
            raise err

        except Exception:
            raise Error.RuntimeError(traceback.format_exc())

    return result



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The GatePulsePair object.
"""

import sys
import copy
from typing import List, Union, Optional

from Quanlse.QOperation.FixedGate import FixedGateOP
from Quanlse.QOperation.RotationGate import RotationGateOP
from Quanlse.QWaveform import QJob
from Quanlse.QOperation import Error, CircuitLine
from Quanlse.QOperator import QOperator


class GatePulsePair:
    """
    A gate with its corresponding pulses.

    :param cirLine: quantum gate in the quantum circuit model.
    :param pulses: QJob object with corresponding pulses.
    """

    def __init__(self, cirLine: CircuitLine, pulses: QJob = None):
        """ Initialization """
        # Set properties
        self._pulses = None  # type: Optional[QJob]
        self._cirLine = None  # type: Optional[CircuitLine]
        self._onSubSys = []  # type: List[int]
        self._t = 0.0  # type: float

        # Set objects
        self.cirLine = copy.deepcopy(cirLine)
        self.pulses = copy.deepcopy(pulses)

    def __eq__(self, other):
        """ Operator == """
        return self._compareObj(other)

    def __ne__(self, other):
        """ Operator != """
        return not self._compareObj(other)

    @property
    def pulses(self) -> Union[QJob, None]:
        """
        Get pulses property
        """
        return self._pulses

    @pulses.setter
    def pulses(self, item: Optional[QJob]):
        """
        Set pulses property
        """
        if item is not None:
            self._pulses = copy.deepcopy(item)
            # pulse duration
            self._t, _ = self.pulses.computeMaxTime()
            # Generate onSubSys
            onSubSys = []
            for opKey in self.pulses.ctrlOperators.keys():
                ops = self.pulses.ctrlOperators[opKey]
                if isinstance(ops, list):
                    for op in ops:
                        onSubSys.append(op.onSubSys)
                elif isinstance(ops, QOperator):
                    onSubSys.append(ops.onSubSys)

            onSubSys = list(set(onSubSys) | set(self.cirLine.qRegIndexList))
            onSubSys.sort()
            self._onSubSys = onSubSys

    @property
    def t(self) -> float:
        """
        Pulse duration
        """
        return self._t

    @property
    def onSubSys(self) -> List[int]:
        """
        The sub-systems which the pulses work on.
        """
        return self._onSubSys

    @property
    def cirLine(self):
        """
        Get cirLine property
        """
        return self._cirLine

    @cirLine.setter
    def cirLine(self, item):
        """
        Set cirLine property
        """
        self._cirLine = copy.deepcopy(item)

    def _compareObj(self, other):
        """
        Compare two GatePulsePair object.
        """
        if not isinstance(other, GatePulsePair):
            raise Error.ArgumentError(f"GatePulsePair object can not compare with a {type(other)}.")
        if isinstance(other.cirLine.data, FixedGateOP) and isinstance(self.cirLine.data, RotationGateOP):
            return False
        if isinstance(other.cirLine.data, RotationGateOP) and isinstance(self.cirLine.data, FixedGateOP):
            return False
        if other.cirLine.data.name != self.cirLine.data.name:
            return False
        if other.cirLine.qRegIndexList != self.cirLine.qRegIndexList:
            return False
        if isinstance(self.cirLine.data, RotationGateOP):
            argLen = len(other.cirLine.data.uGateArgumentList)
            for idx in range(argLen):
                verify = abs(other.cirLine.data.uGateArgumentList[idx] -
                             self.cirLine.data.uGateArgumentList[idx])
                if verify > sys.float_info.epsilon:
                    return False
        return True

    def hasPulse(self):
        """
        check if the instance is empty
        """
        return True if self.pulses is not None else False


#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quantum Register Pool
"""
from typing import Dict, TYPE_CHECKING


class QRegStorage:
    """
    The storage for quantum register.
    """

    def __init__(self, index: int, env: 'Scheduler') -> None:
        """
        The quantum register object needs to have an index and a quantum environment.

        :param index: the quantum register index
        :param env: quantum environment or procedure
        :return: None
        """

        self.index = index
        self.env = env


class QRegPool:
    """
    The quantum register dict
    """

    def __init__(self, env: 'Scheduler') -> None:
        """
        The constructor of the QRegPool class

        :param env: the related quantum environment or procedure
        :return: None
        """

        # the quantum environment related with the quantum register dict
        self.env = env
        # the inner data for quantum register dict
        self.registerMap = {}  # type: Dict[int, QRegStorage]

    def __getitem__(self, index: int) -> QRegStorage:
        """
        Return the quantum register at the given index

        :param index: item index
        :return: the returned item
        """
        return self._get(index)

    def __call__(self, index: int) -> QRegStorage:
        """
        Return the quantum register at the given index

        :param index: item index
        :return: the returned quantum register
        """
        return self._get(index)

    def _get(self, index: int) -> QRegStorage:
        """
        Get the quantum register according to the index

        Create the register when it does not exists

        :param index: the quantum register index
        :return: QuantumRegisterStorage
        """

        value = self.registerMap.get(index)
        if value is None:
            value = QRegStorage(index, self.env)
            self.registerMap[index] = value
        return value

    def changeEnv(self, env: 'Scheduler') -> None:
        """
        Change the environment

        :param env: the environment to be changed to
        :return: None
        """
        self.env = env
        for qReg in self.registerMap.values():
            qReg.env = env



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quanlse QOperator class.
"""

import pickle
import base64
from math import sqrt
from numpy import ndarray, array, zeros, conjugate, transpose, dot
from typing import Union

from Quanlse.QPlatform import Error


class QOperator:
    """
    In Quanlse, operators are stored in QOperator objects, which keep track of the
    operator's purpose, matrix form, coefficients and the qubits it corresponds to.

    :param name: a user-given name
    :param matrix: corresponding matrix of the operator
    :param onSubSys: subsystem number
    :param coef: corresponding coefficient
    """
    def __init__(self, name: str, matrix: ndarray = None, onSubSys: int = None,
                 coef: Union[float, complex] = 1.0) -> None:
        """
        The constructor of the QOperator class.
        """
        self.name = name  # type: str
        self.matrix = matrix  # type: ndarray
        self.onSubSys = onSubSys  # type: int
        self.coef = coef  # type: float

    @property
    def matrix(self):
        """
        Return the corresponding matrix (type: ndarray)
        """
        return self._matrix

    @matrix.setter
    def matrix(self, matrix: ndarray = None):
        """
        Setter function for the matrix

        :param matrix: the matrix to be set to
        """
        if matrix is not None and not isinstance(matrix, ndarray):
            raise Error.ArgumentError("matrix must be a numpy.ndarray!")
        self._matrix = matrix

    @property
    def name(self):
        """
        Return the corresponding name of the operator.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """
        Modify the name of the operator.

        :param name: name to be changed to
        """
        if not isinstance(name, str):
            raise Error.ArgumentError("name must be a str!")
        self._name = name

    @property
    def onSubSys(self):
        """
        Return the corresponding subsystem number of the operator.
        """
        return self._onSubSys

    @onSubSys.setter
    def onSubSys(self, onSubSys: int):
        """
        Modify the corresponding subsystem of the operator.

        :param onSubSys: subsystem to be changed to
        """
        if onSubSys is None:
            self._onSubSys = None
        else:
            if not isinstance(onSubSys, int):
                raise Error.ArgumentError(f"onSubSys must be an integer, instead of {type(onSubSys)}!")
            self._onSubSys = onSubSys

    @property
    def coef(self):
        """
        Coefficient of the operator
        """
        return self._coef

    @coef.setter
    def coef(self, coef: Union[float, complex] = 1.0):
        """
        Setter for the coefficient

        :param: coefficient to be set to
        """
        if coef is None:
            self._coef = 1.0
        else:
            if not (isinstance(coef, float) or isinstance(coef, complex)):
                raise Error.ArgumentError(f"coef must be a float or complex, instead of {type(coef)}!")
            self._coef = coef

    def dump(self) -> str:
        """
        Return base64 encoded string.

        :return: a base64 encoded string
        """
        # Dump the object
        byteStr = pickle.dumps(self)
        base64str = base64.b64encode(byteStr)
        return base64str.decode()

    @staticmethod
    def load(base64Str: str) -> 'QOperator':
        """
        Create object from base64 encoded string.

        :return: a QOperator object
        """
        byteStr = base64.b64decode(base64Str.encode())
        obj = pickle.loads(byteStr)  # type: QOperator
        return obj


def sigmaI(arg=None) -> QOperator:
    r"""
    Matrix form of the Pauli-I operator:

    :math:`\hat{\sigma}_I = \begin{bmatrix} 1 & 0 \\ 0 & 1
    \end{bmatrix}`

    :return: a QOperator object
    """
    qo = QOperator("sigmaI", matrix=array([
        [1, 0],
        [0, 1]
    ], dtype=complex))
    return qo


def sigmaX(arg=None) -> QOperator:
    r"""
    Matrix form of the Pauli-X operator:

    :math:`\hat{\sigma}_X = \begin{bmatrix} 0 & 1 \\ 1 & 0
    \end{bmatrix}`

    :return: a QOperator object
    """
    qo = QOperator("sigmaX", matrix=array([
        [0, 1],
        [1, 0]
    ], dtype=complex))
    return qo


def sigmaY(arg=None) -> QOperator:
    r"""
    Matrix form of the Pauli-Y operator:

    :math:`\hat{\sigma}_Y = \begin{bmatrix} 0 & -i \\ i & 0
    \end{bmatrix}`

    :return: a QOperator object
    """
    qo = QOperator("sigmaY", matrix=array([
        [0, -1j],
        [1j, 0]
    ], dtype=complex))
    return qo


def sigmaZ(_=None) -> QOperator:
    r"""
    Matrix form of the Pauli-Z operator:

    :math:`\hat{\sigma}_Z = \begin{bmatrix} 1 & 0 \\ 0 & -1
    \end{bmatrix}`

    :return: a QOperator object
    """
    qo = QOperator("sigmaZ", matrix=array([
        [1, 0],
        [0, -1]
    ], dtype=complex))
    return qo


def number(d: int = 2) -> QOperator:
    r"""
    Return the matrix form of the number operator:
    :math:`\hat{a}^{\dagger}\hat{a}`.

    :param d: the dimension of the number operator
    :return: matrix form of the number operator
    """
    return QOperator("number", dot(adagger(d), a(d)))


def uWave(d: int = 2) -> QOperator:
    r"""
    Return the matrix of a microwave driven operator: :math:`(\hat{a} + \hat{a}^{\dagger})`.

    :param d: the dimension of the microwave driven operator
    :return: matrix of form the microwave driven operator
    """
    return QOperator("uWave", a(d) + adagger(d))


def flux(d: int = 2) -> QOperator:
    r"""
    Return the matrix of a magnetic flux driven operator: :math:`\hat{a}^{\dagger}\hat{a}`.

    :param d: the dimension of the magnetic flux channel driven operator
    :return: matrix of form the magnetic flux channel driven operator
    """
    return QOperator("flux", dot(adagger(d), a(d)))


def driveX(d: int = 2) -> QOperator:
    r"""
    Return the matrix of a X-channel driven operator: :math:`(\hat{a} + \hat{a}^{\dagger}) / 2`.

    :param d: the dimension of the X-channel driven operator
    :return: matrix of form the X-channel driven operator
    """
    return QOperator("driveX", (a(d) + adagger(d)) / 2)


def driveY(d: int = 2) -> QOperator:
    r"""
    Return the matrix of a Y-channel driven operator: :math:`i (\hat{a} - \hat{a}^{\dagger}) / 2`.

    :param d: the dimension of the Y-channel driven operator
    :return: matrix form of the Y-channel driven operator
    """
    return QOperator("driveY", 1j * (a(d) - adagger(d)) / 2)


def driveZ(d: int = 2) -> QOperator:
    r"""
    Return the matrix of a Z-channel driven operator: :math:`\hat{a}^{\dagger}\hat{a}`.

    :param d: the dimension of the Y-channel driven operator
    :return: matrix form of the Y-channel driven operator
    """
    return QOperator("driveZ", dot(adagger(d), a(d)))


def duff(d: int = 2) -> QOperator:
    r"""
    Return the matrix form of the Duffing operator:
    :math:`\hat{a}^{\dagger}\hat{a}^{\dagger}\hat{a}\hat{a}`.

    :param d: the dimension of the Duffing operator
    :return: matrix form of the Duffing operator
    """
    mat = dot(dot(adagger(d), adagger(d)), dot(a(d), a(d)))
    return QOperator("duff", mat)


def destroy(d: int = 2) -> QOperator:
    r"""
    Return the matrix form of an annihilation operator :math:`\hat{a}`
    with a dimension of :math:`d` (:math:`d` is an integer):

    :math:`\hat{a} = \begin{bmatrix} 0 & \sqrt{1} & 0 & \cdots & 0 \\ 0 & 0 & \sqrt{2} & \cdots & 0 \\
    \vdots & \vdots & \vdots & \ddots & \vdots \\ 0 & 0 & 0 & \cdots & \sqrt{d-1} \\
    0 & 0 & 0 & \cdots & 0
    \end{bmatrix}`

    :param d: the dimension of the annihilation operator
    :return: matrix form of the annihilation operator
    """
    return QOperator("a", a(d))


def create(d: int = 2) -> QOperator:
    r"""
    Return the matrix form of a creation operator :math:`\hat{a}^{\dagger}`
    with a dimension of :math:`d` (:math:`d` is an integer):

    :math:`\hat{a}^{\dagger} = \begin{bmatrix} 0 & 0 & \cdots & 0 & 0 \\ \sqrt{1} & 0 & \cdots & 0 & 0 \\
    0 & \sqrt{2} & \cdots & 0 & 0 \\ \vdots & \vdots & \ddots & \vdots & \vdots \\
    0 & 0 & \cdots & \sqrt{d-1} & 0
    \end{bmatrix}`

    :param d: the dimension of the creation operator
    :return: matrix form of the creation operator
    """
    return QOperator("adag", adagger(d))


def dagger(matrix: ndarray) -> ndarray:
    """
    Return the conjugate transpose of a given matrix.

    :param matrix: the given matrix
    :return: the conjugate transposed matrix
    """
    return array(conjugate(transpose(matrix)), order="C")


def a(d: int = 2) -> ndarray:
    """
    Return the numpy matrix of destroy operator.

    :return: a numpy matrix
    """
    mat = zeros((d, d), dtype=complex, order="C")
    for i in range(0, d - 1):
        mat[i, i + 1] = sqrt(i + 1) + 0 * 1j
    return mat


def adagger(d: int = 2) -> ndarray:
    """
    Return the numpy matrix of creation operator.

    :return: a numpy matrix
    """
    return dagger(a(d))



# !/usr/bin/python3
# -*- coding: utf8 -*-
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions for plotting in ErrorMitigation module.
"""

import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
from typing import List

from Quanlse.QHamiltonian import QHamiltonian
from Quanlse.QOperation import FixedGate
from Quanlse.Utils.Functions import expect
from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.Define import outputPath


def polarVectorToState(vec) -> np.ndarray:
    """
    From the polar vector representation to the state representation.

    :param vec: polar vector, array_like, (3,)
    :return: density matrix: ndarray, (2, 2)
    """
    ops = [FixedGate.X.getMatrix(), FixedGate.Y.getMatrix(), FixedGate.Z.getMatrix()]
    return (np.identity(2) + vec[0] * ops[0] + vec[1] * ops[1] + vec[2] * ops[2]) / 2


def stateToPolarVector(state) -> List:
    """
    From the state representation to the polar vector representation.

    :param state: state vector or density matrix
    :return: polar vector
    """
    ops = [FixedGate.X.getMatrix(), FixedGate.Y.getMatrix(), FixedGate.Z.getMatrix()]
    return [expect(ops[i], state) for i in range(3)]


def plotZNESequences(expectationsRescaled: List[List], expectationsExtrapolated: List[List], expectationsIdeal: List,
                     fileName: str = None):
    """
    Plot a figure that shows the extrapolated results and noise-rescaling results, with the X-axis
    representing the size of quantum gate sequences and the Y-axis representing the expectation
    values of the given mechanical quantity.

    :param expectationsRescaled: a series of noise-rescaling expectation values with shape [n, d + 1]
    :param expectationsExtrapolated: a series of extrapolated expectation values with shape [n, d]
    :param expectationsIdeal: a series of ideal expectation values with shape [n,]
    :param fileName: file name
    :return: None
    """
    # shape: [numSeq, d + 1] --> [d + 1, numSeq]
    expectationsRescaled = np.array(expectationsRescaled).transpose()
    # shape: [numSeq, d] --> [d, numSeq]
    expectationsExtrapolated = np.array(expectationsExtrapolated).transpose()
    numSeq = expectationsRescaled.shape[1]
    lengthEachSequence = list(range(1, numSeq + 1))

    # plot results
    plt.figure(figsize=(10, 6))
    for d, Es in enumerate(expectationsRescaled):
        plt.plot(lengthEachSequence, Es, 'o-', label='{}-order rescaled'.format(d + 1))
    for d, Es in enumerate(expectationsExtrapolated):
        plt.plot(lengthEachSequence, Es, 'o--', label='{}-order extrapolated'.format(d + 1))
    plt.plot(lengthEachSequence, expectationsIdeal, 'k+--', label='Ideal')
    plt.xlabel('Size of Clifford sequence')
    plt.ylabel('Expectation value')
    plt.title('Richardson Extrapolation Result of Different Order')
    plt.legend()
    plt.xticks(lengthEachSequence)
    if fileName is not None:
        localFile = os.path.join(outputPath, fileName)
        plt.savefig(localFile)
    plt.show()


def plotRescaleHamiltonianPulse(rescaleCoes: List[float], hamList: List[QHamiltonian], numChannel: int,
                                figsize: tuple = (12, 6), title=None):
    """
    Plot each pulse channel of a series of time-rescaling Hamiltonians. Suppose len(rescaleCoes)=r and numChannel=n,
    this function will show a figure including r*c subfigures which are
    arranged in "r" rows and "c" columns.

    :param rescaleCoes: rescaling coefficients
    :param hamList: Hamiltonian list whose length is equal to `rescaleCoes`
    :param numChannel: number of pulse channels
    :param figsize: size of the entire figure
    :param title: super title of the entire figure
    :return: fig, axes
    """
    fig = plt.figure(figsize=figsize)
    axes = []
    colorList = plt.rcParams['axes.prop_cycle'].by_key()['color']

    if len(rescaleCoes) != len(hamList):
        raise ArgumentError("Length of 'rescaleCoes' and length of 'hamList' are not equal!")
    grids = gridspec.GridSpec(len(rescaleCoes), numChannel)
    # dtList: e.g. [1, 1.25, ...]
    dtList = [ham.dt for ham in hamList]
    # tList: e.g. [1*16, 1.25*16, ...]
    tList = [ham.dt * ham.job.endTimeDt for ham in hamList]

    tMax = max(tList)
    driveNames = ['Drive X Pulse', 'Drive Y Pulse']
    if numChannel == 3:
        driveNames.append('Drive Z Pulse')

    for c, dr in enumerate(driveNames):
        for r, coe in enumerate(rescaleCoes):
            dt = dtList[r]
            ham = hamList[r]

            # ax lies in (row, col): (r, c)
            ax = plt.subplot(grids[r, c])

            # key: the name of pulse channel
            key = list(ham.ctrlCache.keys())[c]
            endDt = ham.job.endTimeDt
            As = [ham.job.waveCache[key][i] for i in range(endDt)]
            tBar = []
            Abar = []
            # construct tBar (list type) and Abar (list type) to plot a bar figure
            for i in range(endDt + 1):
                tBar.extend([i * dt, i * dt])
                if i == 0:
                    Abar.extend([0, As[i]])
                elif i == endDt:
                    Abar.extend([As[i - 1], 0])
                else:
                    Abar.extend([As[i - 1], As[i]])

            plt.fill(tBar, Abar, color=colorList[r], alpha=0.6)
            plt.plot(tBar, Abar, color=colorList[r], label='coe: {:.2f}'.format(coe))
            plt.legend(loc='upper right')

            plt.xlim(0, tMax)

            if r == 0:
                plt.title(key)
            if r == len(rescaleCoes) - 1:
                plt.xlabel('Times')

            axes.append(ax)

    fig.tight_layout()
    plt.suptitle(title)
    return fig, axes



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Quanlse's Scheduler Superconduct achieves the following goals:

- It generates parameters and AWG input signal arrays for fidelity-optimized pulses
  when leakage into the state :math:`|2\rangle` is taken into account.

- It is also capable of scheduling pulses to minimize idle time and therefore reduce
  decoherence losses.

- At the same time, it allows for the visualization of pulse sequences for the users
  to check the results.

CNOT gate is rarely directly implemented on superconducting quantum chips. Instead,
it is often constructed by piecing together single-qubit gates and other two-qubit
gates like CR gate or ISWAP gate that can be easily implemented on a superconducting
chip (often called native gates). The two-qubit gates that are available in the
transmon-like superconducting qubit architecture can be divided into two categories:

**Flux-controlled**

This class of gates offers the advantage of short gate time to minimize decoherence
error. However, tuning the qubit frequency can introduce flux noises and lead to the
problem of frequency crowding.

**All-microwave control**

CR gate allows for an all-microwave control, which alleviates the problem of flux
noise. However, the much longer time-scale limits the gate fidelity (because of
the decoherence effects of qubits).

Since CZ gates can be used to construct a CNOT gate easily by using only two other
single-qubit gates, Quanlse's Scheduler offers this way to construct a CNOT gate
in a quantum circuit.
"""
import copy
from typing import List

from Quanlse.Scheduler import Scheduler
from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.Scheduler.SchedulerPulseGenerator import SchedulerPulseGenerator
from Quanlse.Scheduler.SchedulerPipeline import SchedulerPipeline


class SchedulerSuperconduct(Scheduler):
    """
    Basic class of Quanlse Scheduler for superconducting platform

    :param dt: AWG sampling time
    :param ham: the QHamiltonian object.
    :param generator: the pulseGenerator object.
    :param subSysNum: size of the subsystem
    :param sysLevel: the energy levels of the system (support different levels for different qubits)
    """
    def __init__(self, dt: float = None, ham: QHam = None, generator: SchedulerPulseGenerator = None,
                 pipeline: SchedulerPipeline = None, subSysNum: int = None, sysLevel: int = None):
        """
        Constructor for SchedulerSuperconduct object
        """
        # Initialization
        super().__init__(dt, ham, generator, pipeline, subSysNum, sysLevel)

        # Add the pulse generator
        if self._pulseGenerator is None:
            if ham is not None:
                from Quanlse.Scheduler.Superconduct.GeneratorCloud import generatorCloud
                self._pulseGenerator = generatorCloud(self._ham)
        else:
            self._pulseGenerator = copy.deepcopy(generator)

        # Add the pipeline
        if pipeline is not None:
            del self._pipeline
            self._pipeline = copy.deepcopy(pipeline)
        else:
            from Quanlse.Scheduler.Superconduct.PipelineLeftAligned import leftAligned
            self.pipeline.addPipelineJob(leftAligned)

        # Set the local oscillator information
        self._subSysLO = []  # type: List[List[float]]
        for num in range(self.subSysNum):
            self._subSysLO.append([0.0, 0.0])



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions used in ErrorMitigation module.
"""

import numpy as np
import matplotlib.pyplot as plt
from math import pi
from itertools import combinations
from scipy import linalg
from typing import List

from Quanlse.QOperator import number, duff
from Quanlse.QHamiltonian import QHamiltonian
from Quanlse.QOperation import RotationGate, CircuitLine
from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.Scheduler.Superconduct import SchedulerSuperconduct
from Quanlse.Utils.Functions import project, tensor, expect, fromMatrixToAngles
from Quanlse.Utils.Plot import plotPulse
from Quanlse.Utils.Clifford import randomClifford
from Quanlse.Scheduler.Superconduct.GeneratorRBPulse import SingleQubitCliffordPulseGenerator


def setupBasicHamiltonian(n: int = None) -> QHamiltonian:
    r"""
    Setup the system's Hamiltonian with prefixed parameters.

    :param n: number of qubits of the Hamiltonian
    :return: the basic Hamiltonian
    """
    if n is None:
        raise ArgumentError("Error in setupBasicHamiltonian(): number of qubits is not specified!")
    # Prefix the coupling strength of qubit pairs
    g = 0.0038 * 2 * pi
    # Anharmonicity
    anharm = -0.33 * 2 * pi
    # Qubit frequency list
    qubitFreq = [4.914 * 2 * pi,
                 5.114 * 2 * pi,
                 4.914 * 2 * pi,
                 5.114 * 2 * pi]
    # Set the rotating wave approximation frequency
    rwa = 4.914 * 2 * pi
    # Set the unit time for solving ODE
    dt = 1.0
    # Define the system Hamiltonian
    ham = QHamiltonian(subSysNum=n, sysLevel=3, dt=dt)
    # Add drift terms
    for q in range(n):
        ham.addDrift(number(3), q, qubitFreq[q] - rwa)
        ham.addDrift(duff(3), q, anharm)
    # Add coupling terms. If there is only one qubit, there is no coupling term.
    # We assume `full connectivity` of the quantum device,
    # thus there is a coupling term between arbitrary two qubits.
    for cq, tq in combinations(range(n), 2):
        ham.addCoupling([cq, tq], g)
    return ham


def fromCircuitToHamiltonian(circuit: List[CircuitLine] = None, useCliffordPulse: bool = True) -> QHamiltonian:
    r"""
    Convert a quantum circuit into the optimal Hamiltonian.

    :param circuit: the quantum circuit to be optimized. It is a list of `CircuitLine` objects, each `CircuitLine`
            specifies a quantum gate operating on several qubits
    :param useCliffordPulse: indicator for using the built-in Clifford pulse or not
    :return: The optimal Hamiltonian which is the optimal implementation of the unitary evolution of the gate sequence
    """
    # Step 1. Setup the basic Hamiltonian with prefixed parameters. These parameters are from the quantum devices.
    # Parse the number of qubits used in the quantum circuit
    n = numberOfQubitsOfCircuit(circuit)
    # Construct system Hamiltonian with basic information from quantum device
    ham = setupBasicHamiltonian(n)

    # Step 2. Use the Quanlse Scheduler to generate pulse sequences for the input quantum circuit
    if useCliffordPulse:
        scheduler = SchedulerSuperconduct(ham.dt, ham, generator=SingleQubitCliffordPulseGenerator(ham))
    else:
        scheduler = SchedulerSuperconduct(ham.dt, ham)
    # Translate the circuits to gates operated on the Scheduler
    for gate in circuit:
        if len(gate.qRegIndexList) == 1:
            gate.data(scheduler.Q[gate.qRegIndexList[0]])
        elif len(gate.qRegIndexList) == 2:
            gate.data(scheduler.Q[gate.qRegIndexList[0]], scheduler.Q[gate.qRegIndexList[1]])
        else:
            raise ArgumentError('Error in fromCircuitToHamiltonian(): unrecognized gate type!')

    # Run the scheduler
    scheduler.schedule()
    # Build the cache for the Hamiltonian
    scheduler.ham.buildCache()
    # Return the optimized Hamiltonian corresponding to the quantum circuit
    return scheduler.ham


def fromHamiltonianToOperator(ham: QHamiltonian) -> np.ndarray:
    r"""
    Call the `simulate` function to compute the corresponding evolution operator of the given Hamiltonian.

    :param ham: the Hamiltonian, with evolution time and wave sequences given
    :return: The evolution operator
    """
    result = ham.simulate(refreshCache=False)
    projectedEvolution = project(result.result[0]["unitary"], ham.subSysNum, ham.sysLevel, 2)
    return projectedEvolution


def numberOfQubitsOfCircuit(circuit: List[CircuitLine] = None) -> int:
    r"""
    Extract the number of qubits involved in the given quantum circuit.
    We implement this by collecting the qubit indices and count the number.

    :param circuit: the quantum circuit which is a list of `CircuitLine` objects.
    :return: number of qubits used in this quantum circuit
    """
    qubitIndices = []
    for gate in circuit:
        qubitIndices.extend(gate.qRegIndexList)
    qubitIndices = np.unique(qubitIndices)
    return len(qubitIndices)


def printHamiltonian(ham: QHamiltonian) -> None:
    r"""
    Print the basic information of a Hamiltonian object.
    This function is used within the ZNE module for illustrative purpose.

    :param ham: the given Hamiltonian
    :return: None
    """
    print('=================================================================')
    print('The basic information of the Hamiltonian is:')
    print(ham)
    print('Print the control pulses now ...')
    print("ham.job.endTimeDt = ", ham.job.endTimeDt)
    print(ham.ctrlCache)
    for key in ham.ctrlCache:
        for nowDt in range(0, ham.job.endTimeDt):
            print("ham.job.waveCache[{}][{}] = {}".format(key, nowDt, ham.job.waveCache[key][nowDt]))


def computeInverseGate(circuit: List[CircuitLine]) -> CircuitLine:
    r"""
    Compute the inverse gate of a sequence of single-qubit gates.
    Notice that a decomposition of single-qubit gates can be characterized via:

    :math:`U = e^{i\alpha} R_z(\phi) R_y(\theta) R_z(\lambda)
    = e^{i(\alpha-\phi/2-\lambda/2)}
    \begin{bmatrix}
    \cos(\theta/2) & -e^{i\lambda}\sin(\theta/2) \\
    e^{i\phi}\sin(\theta/2) & e^{i(\phi+\lambda)}\cos(\theta/2)
    \end{bmatrix}
    = e^{i(\alpha-\phi/2-\lambda/2)} U'`

    We define the inverse to be the inverse of U'. That is,
    we do not consider the effect of global phases.
    Example:
        The input is `[X, Y, Z, H]`, the output will be `XYZH`

    :param circuit: the quantum circuit whose inverse to be computed.
            It is a list of `CircuitLine` objects, each `CircuitLine` specifies a quantum gate operating on several qubits.
    :return: the inverse gate in type of `CircuitLine`
    """
    if len(circuit) < 1:
        raise ArgumentError(f"The input gate sequence should contain at least one gate!")

    n = numberOfQubitsOfCircuit(circuit)
    product = np.identity(2 ** n)
    for item in circuit:
        product = item.data.getMatrix() @ product
    # Compute the inverse gate of the composite Clifford gates
    invMatrix = np.linalg.inv(product)
    # Transform the inverse gate to RotationGate.U format and add it to the circuit
    [_, theta, phi, lamda] = fromMatrixToAngles(invMatrix)
    return CircuitLine(RotationGate.U(theta, phi, lamda), [0])


def computeIdealEvolutionOperator(circuit: List[CircuitLine]) -> np.ndarray:
    """
    Compute the evolution unitary operator of the given circuit.

    :param circuit: the quantum circuit whose corresponding ideal evolution to be computed.
            It is a list of `CircuitLine` objects, each `CircuitLine` specifies a quantum gate operating on several qubits.
    :return: ndarray-type with shape (2^n, 2^n)
    """
    n = numberOfQubitsOfCircuit(circuit)
    U = tensor([np.identity(2)] * n)
    for gate in circuit:
        if len(gate.qRegIndexList) == 1:
            # Form a list of identities I \otimes ... \otimes I
            V = tensorWithSingleQubitGate(gate.data.getMatrix(), n, gate.qRegIndexList[0])
            U = V @ U
        elif len(gate.qRegIndexList) == 2:
            V = tensorWithTwoQubitGate(gate.data.getMatrix(), n, gate.qRegIndexList[0], gate.qRegIndexList[1])
            U = V @ U
        else:
            raise ArgumentError(
                'Error in computeIdealEvolutionOperator(): only single and two-qubit gates are supported!')
    return U


def computeIdealExpectationValue(rho: np.ndarray,
                                 circuit: List[CircuitLine] = None,
                                 A: np.ndarray = None) -> float:
    r"""
    Compute the ideal expectation value of the following form

    :math:`{\rm{Tr}}[A C_n\cdots C_1 \rho C_1^\dagger \cdots C_n^\dagger]`

    where each :math:`C_i` is a quantum gate.

    :param rho: the input quantum state
    :param circuit: the quantum circuit. It is a list of `CircuitLine` objects,
            each `CircuitLine` specifies a quantum gate operating on several qubits.
    :param A: the quantum observable
    :return: the ideal expectation value
    """
    n = numberOfQubitsOfCircuit(circuit)
    if (rho.shape[0] != 2 ** n) or (rho.shape[1] != 2 ** n):
        raise ArgumentError("the input state must have the same dimension of the quantum circuit!")
    if (A.shape[0] != 2 ** n) or (A.shape[1] != 2 ** n):
        raise ArgumentError("the observable must have the same dimension of the quantum circuit!")

    U = computeIdealEvolutionOperator(circuit)
    return expect(A, U @ rho @ U.conj().T)


def tensorWithSingleQubitGate(singleQubitOperator: np.ndarray, n: int, qubitIndex: int) -> np.ndarray:
    """
    Given a single-qubit gate, we compute its tensor unitary matrix expanded
    to the whole Hilbert space (totally n qubits).

    :param singleQubitOperator: ndarray with shape (2,2)
    :param n: number of qubits
    :param qubitIndex: index of the qubit with the subsequent single-qubit operation
    :return: ndarray with shape (2^n, 2^n)
    """
    if qubitIndex not in range(n):
        raise ArgumentError("The qubitIndex is out of range!")
    qubitGateList = [np.identity(2)] * n
    # Substitute the i-th gate with the given gate
    qubitGateList[qubitIndex] = singleQubitOperator
    # Use tensor function to obtain the composite unitary operator
    U = tensor(qubitGateList)
    return U


def tensorWithTwoQubitGate(twoQubitOperator: np.ndarray, n: int, cq: int, tq: int) -> np.ndarray:
    """
    Much similar to the `tensorWithSingleQubitGate` function,
    this function aims to compute the tensor unitary matrix expanded
    to the whole Hilbert space (totally n qubits) of a two-qubit gate.

    :param twoQubitOperator: ndarray with shape (4,4)
    :param n: number of qubits
    :param cq: control qubit index
    :param tq: target qubit index
    :return: ndarray with shape (2^n, 2^n)
    """
    if cq not in range(n) or tq not in range(n):
        raise ArgumentError("The control or the target qubit is out of range!")

    qubitGateList = [np.identity(2)] * (n - 1)
    qubitGateList[0] = twoQubitOperator
    U = tensor(qubitGateList).reshape([2] * 2 * n)
    idx = np.repeat(-1, n)
    idx[cq] = 0
    idx[tq] = 1
    idx[idx < 0] = range(2, n)
    idx = idx.tolist()
    idxLatter = [i + n for i in idx]
    U = np.transpose(U, idx + idxLatter).reshape([2 ** n, 2 ** n])
    return U


def plotHamiltonianPulse(ham: QHamiltonian) -> None:
    r"""
    Print the pulse of a Hamiltonian object.

    :param ham: the given Hamiltonian
    :return: None
    """
    if ham.ctrlCache is None or len(ham.ctrlCache) == 0:
        raise ArgumentError("in plotHamiltonianPulse(): the control cache is not set!")
    # Create two empty lists
    x = []
    y = []
    yLabel = []
    colors = []
    for key in ham.ctrlCache:
        aList = np.array(ham.waveCache[key])
        tList = np.linspace(0, len(aList) * ham.dt, num=len(aList))
        y.append(list(aList))
        x.append(list(tList))
        yLabel.append('Amp (a.u.)')
        colors.append('blue')
    plotPulse(x, y, xLabel=f'Time (ns)', yLabel=yLabel, title=list(ham.ctrlCache), color=colors, dark=False)
    plt.show()


def globalPhase(U: np.ndarray) -> float:
    r"""
    Compute the global phase of a 2*2 unitary matrix.
    Each 2*2 unitary matrix can be equivalently characterized as:

    :math:`U = e^{i\alpha} R_z(\phi) R_y(\theta) R_z(\lambda)`

    We aim to compute the global phase `\alpha`.
    See also Theorem 4.1 in `Nielsen & Chuang`'s book.

    :param U: the matrix representation of the 2*2 unitary
    :return: the global phase of the unitary matrix
    """
    # Notice that the determinant of the unitary is given by e^{2i\alpha}
    coe = linalg.det(U) ** (-0.5)
    alpha = - np.angle(coe)
    return alpha


def removeGlobalPhase(U: np.ndarray) -> np.ndarray:
    r"""
    Remove the global phase of a 2*2 unitary matrix.
    Each 2*2 unitary matrix can be equivalently characterized as:

    :math:`U = e^{i\alpha} R_z(\phi) R_y(\theta) R_z(\lambda)`

    We aim to remove the global phase `e^{i\alpha}` from the unitary matrix.
    See Theorem 4.1 in `Nielsen & Chuang`'s book for details.

    :param U: the matrix representation of the 2*2 unitary operator
    :return: the unitary matrix whose global phase has been removed
    """
    alpha = globalPhase(U)
    U = U * np.exp(- 1j * alpha)
    return U


def rotationXZCircuit(numSeq: int = 5) -> List[CircuitLine]:
    r"""
    Return a list including serial X-Z rotation gates, while the whole effect of all elements is equal to a X gate.

    :math:`U_j = R_z(\frac{4j\pi}{n})R_x(\frac{j\pi}{n})R_x(\frac{-(j-1)\pi}{n})R_z(\frac{-4(j-1)\pi}{n})`

    :param numSeq: length of the X-Z rotation gates sequences
    :return: a list including serial X-Z rotation gates ("numSeq" elements in total)
    """
    circuit = []
    n = numSeq
    for k in range(1, numSeq + 1):
        rzp = RotationGate.RZ(4 * k * pi / n).getMatrix()
        rxp = RotationGate.RX(k * pi / n).getMatrix()
        rxm = RotationGate.RX(-(k - 1) * pi / n).getMatrix()
        rzm = RotationGate.RZ(-4 * (k - 1) * pi / n).getMatrix()
        Uk = rzp @ rxp @ rxm @ rzm
        _, theta, phi, lamda = fromMatrixToAngles(Uk)
        circuit.append(CircuitLine(RotationGate.U(theta, phi, lamda), [0]))
    return circuit


def randomCircuit(qubits: int = 1, numSeq: int = 5, seed: int = None, type: str = 'clifford') -> List[CircuitLine]:
    r"""
    Randomly generate a numSeq-length quantum circuit based on `size` number of Clifford operators,
    whose number of qubits is given by `qubits`. Currently only single-qubit circuit is supported.

    :param qubits: number of qubits
    :param numSeq: number of random Clifford gates
    :param seed: random number seed
    :param type: 'clifford' or other optional types
    :return: a circuit (`list` type) including a series of gates (`CircuitLine` type)
    """
    if type != 'clifford':
        raise ArgumentError("This function currently can only generate random gates of Clifford type!")
    cliffordGates = randomClifford(qubits, numSeq, seed=seed)
    circuit = [CircuitLine(gate, [0]) for gate in cliffordGates]

    return circuit



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pulse generator for Quanlse Scheduler.
"""

from typing import Dict, Any, List, Callable

from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.QWaveform import QJob
from Quanlse.QOperation import Error, CircuitLine


class SchedulerPulseGenerator:
    """
    Basic class of pulse Generator for Quanlse Scheduler.

    :param ham: a user-given title to the pi object
    """
    def __init__(self, ham: QHam = None):
        self._ham = ham  # type: QHam
        self._generators = {}  # type: Dict[str, Callable]
        self._conf = {}  # type: Dict[str, Any]

    def __call__(self, *args, **kwargs) -> QJob:
        """
        Call the instance and generate pulses.
        """
        if not isinstance(args[0], CircuitLine):
            raise Error.ArgumentError("You should input a circuitLine instance.")
        _cirLine = args[0]  # type: CircuitLine
        _scheduler = args[1]  # type: 'Scheduler'
        # Call the generator according to the input CircuitLine instance, and
        #     input the system Hamiltonian and the configurations.
        if self._ham is None:
            return self[_cirLine.data.name](cirLine=_cirLine, scheduler=_scheduler)
        else:
            return self[_cirLine.data.name](ham=self._ham, cirLine=_cirLine, scheduler=_scheduler)

    def __getitem__(self, gateName: str) -> Callable:
        """
        Get a generator function according to the gateName.
        """
        if gateName in self._generators:
            return self._generators[gateName]
        else:
            raise Error.ArgumentError(f"Gate {gateName} is not supported.")

    @property
    def generators(self) -> Dict[str, Callable]:
        return self._generators

    @property
    def conf(self):
        return self._conf

    def addGenerator(self, gateList: List[str], generator: Callable) -> None:
        """
        Add pulse generator.

        :param gateList: a gate list to add the generator to
        :param generator: generator to be added
        :return: None
        """
        for gateName in gateList:
            self.generators[gateName] = generator



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
In this page, we present the methods for accessing the Quanlse Cloud Service. At present,
we provide services for superconducting platform and ion trap platform.

For more details, please visit:

*Superconduct*

- Superconducting single-qubit gates: https://quanlse.baidu.com/#/doc/tutorial-single-qubit

- GRAPE (GRadient Ascent Pulse Engineering): https://quanlse.baidu.com/#/doc/tutorial-single-qubit

- Superconducting Cross-resonance gates: https://quanlse.baidu.com/#/doc/tutorial-cr

- Superconducting Controlled-Z gates: https://quanlse.baidu.com/#/doc/tutorial-cz

- Superconducting iSWAP gates: https://quanlse.baidu.com/#/doc/tutorial-iswap

*Ion Trap*

- Ion single/two-qubit gates: https://quanlse.baidu.com/#/doc/tutorial-ion-trap-single-and-two-qubit-gate

- Ion Mølmer-Sørensen gates: https://quanlse.baidu.com/#/doc/tutorial-general-MS-gate

"""

from Quanlse.QPlatform.Utilities import (
    numpyMatrixToDictMatrix,
    dictMatrixToNumpyMatrix
)
from Quanlse.QHamiltonian import QHamiltonian as QHam
from Quanlse.QWaveform import QJob

from Quanlse.QRpc import rpcCall

from typing import Dict, Any, List, Tuple, Union
from numpy import ndarray


def remoteOptimize1Qubit(ham: QHam, uGoal: Union[ndarray, List[ndarray]], targetInfid, depth=3) \
        -> Union[Tuple[QJob, float], Tuple[List[QJob], List[float]]]:
    """
    Optimize an arbitrary superconducting single-qubit gate by Quanlse cloud service.

    :param ham: the QHamiltonian object.
    :param uGoal: the goal unitary matrices for optimization.
    :param depth: maximum circuit depth (pulse number).
    :param targetInfid: target infidelity.
    :return: a tuple containing the QJob list and infidelity list.
    """
    args = [ham.dump(), numpyMatrixToDictMatrix(uGoal)]
    kwargs = {"targetInfid": targetInfid, "depth": depth}
    origin = rpcCall("1Qubit", args, kwargs)

    if isinstance(uGoal, list):
        return [QJob.load(o) for o in origin["job"]], origin["inf"]
    else:
        return QJob.load(origin["job"]), origin["inf"]


def remoteOptimize1QubitGRAPE(ham: QHam, uGoal: Union[ndarray, List[ndarray]], tg: int = 20, iterate: int = 150,
                              xyzPulses: List[int] = None) -> Union[Tuple[QJob, float], Tuple[List[QJob], List[float]]]:
    """
    Optimize a single-qubit gate using Gradient Ascent Pulse Engineering.

    :param ham: the QHamiltonian object.
    :param uGoal: the target unitary.
    :param tg: gate time.
    :param iterate: max number of iteration.
    :param xyzPulses: a list of three integers indicating the numbers of
                      Gaussian pulses superposed on x, y, z controls respectively.
    :return: a tuple containing the QJob list and infidelity list.
    """
    args = [ham.dump(), numpyMatrixToDictMatrix(uGoal)]
    kwargs = {
        "tg": tg,
        "iterate": iterate,
        "xyzPulses": xyzPulses
    }
    origin = rpcCall("1QubitGRAPE", args, kwargs)

    if isinstance(uGoal, list):
        return [QJob.load(o) for o in origin["job"]], origin["inf"]
    else:
        return QJob.load(origin["job"]), origin["inf"]


def remoteOptimizeCr(ham: QHam, aBound: Tuple[float, float] = None, tg: float = 200, maxIter: int = 5,
                     targetInfidelity: float = 0.01) -> Tuple[QJob, float]:
    """
    Optimize a superconducting Cross-Resonance gate by Quanlse Cloud Service.

    :param ham: the QHamiltonian object.
    :param aBound: the optimization bound of pulse amplitude.
    :param tg: gate time.
    :param maxIter: max number of iteration.
    :param targetInfidelity: the target infidelity.
    :return: a tuple containing the return Hamiltonian and infidelity.
    """
    args = [ham.dump()]
    kwargs = {
        "aBound": aBound,
        "tg": tg,
        "maxIter": maxIter,
        "targetInfidelity": targetInfidelity
    }
    origin = rpcCall("CR", args, kwargs)
    return QJob.load(origin["job"]), origin["inf"]


def remoteOptimizeCz(ham: QHam, aBound: Tuple[float, float] = None, tg: float = 200, maxIter: int = 5,
                     targetInfidelity: float = 0.01) -> Tuple[QJob, float]:
    """
    Optimize a superconducting Controlled-Z gate by Quanlse Cloud Service.

    :param ham: the QHamiltonian object.
    :param aBound: the optimization bound of pulse amplitude.
    :param tg: gate time.
    :param maxIter: max number of iteration.
    :param targetInfidelity: the target infidelity.
    :return: a tuple containing the return Hamiltonian and infidelity.
    """
    args = [ham.dump()]
    kwargs = {
        "aBound": aBound,
        "tg": tg,
        "maxIter": maxIter,
        "targetInfidelity": targetInfidelity
    }
    origin = rpcCall("CZ", args, kwargs)
    return QJob.load(origin["job"]), origin["inf"]


def remoteOptimizeISWAP(ham: QHam, aBound: Tuple[float, float] = None, tg: float = 200, maxIter: int = 5,
                        targetInfidelity: float = 0.01) -> Tuple[QJob, float]:
    """
    Optimize a superconducting iSWAP gate by Quanlse cloud service.

    :param ham: the QHamiltonian object.
    :param aBound: the optimization bound of pulse amplitude.
    :param tg: gate time.
    :param maxIter: max number of iteration.
    :param targetInfidelity: the target infidelity.
    :return: a tuple containing the return Hamiltonian and infidelity.
    """
    args = [ham.dump()]
    kwargs = {
        "aBound": aBound,
        "tg": tg,
        "maxIter": maxIter,
        "targetInfidelity": targetInfidelity
    }
    origin = rpcCall("ISWAP", args, kwargs)
    return QJob.load(origin["job"]), origin["inf"]


def remoteIonOptimize1Qubit(axial: str, theta: float, tg: float) -> Tuple[float, float, ndarray]:
    """
    Optimize a superconducting iSWAP gate by Quanlse cloud service.

    :param axial: the rotating axial, 'ionRx' or 'ionRy'.
    :param theta: the angle of the rotation operation.
    :param tg: gate time.
    :return: a tuple containing the return Hamiltonian and infidelity.
    """
    args = [axial, theta, tg]
    kwargs = {}
    origin = rpcCall("Ion1Qubit", args, kwargs)
    return origin["a"], origin["b"], dictMatrixToNumpyMatrix(origin["qam"], complex)


def remoteIonMS(ionNumber: int, atomMass: int, tg: float, omega: Tuple[float, float], ionIndex: Tuple[int, int],
                phononMode: str = 'axial', pulseWave: str = 'squareWave') -> Tuple[Any, Any]:
    """
    Generate the Molmer-Sorensen gate in trapped ion

    :param ionNumber: the number of ions.
    :param atomMass: the atomic mass of the ion.
    :param tg: gate time.
    :param omega: 1-dimensional angular frequency of the potential trap.
    :param ionIndex: the index of the two ions.
    :param phononMode: the mode of the phonon oscillation.
    :param pulseWave: the waveform of the laser.
    :return: dict type result and ndarray type unitary.
    """
    args = [ionNumber, atomMass, tg, omega, ionIndex]
    kwargs = {
        "phononMode": phononMode,
        "pulseWave": pulseWave,
    }
    origin = rpcCall("IonMS", args, kwargs)
    return origin['result'], dictMatrixToNumpyMatrix(origin["unitary"], complex)


def remoteIonGeneralMS(gatePair: List[List[int]], args1: Tuple[int, int, float, float, str],
                       args2: Tuple[int, float, float]) -> Tuple[Dict[str, Any], ndarray]:
    """
    Generate general Molmer-Sorensen gate and GHZ state in trapped ion

    :param gatePair: the gate pair in ion chain.
    :param args1:
     args1[0]: the number of ions;
     args1[1]: the atom mass or atom specie;
     args1[2]: the XY trapped potential frequency;
     args1[3]: the Z trapped potential frequency;
     args1[4]: the mode of the phonon oscillation.
    :param args2:
     args2[0]: the pulse sequence number;
     args2[1]: the sample time;
     args2[2]: the laser detuning.
    :return: dict type result and ndarray type unitary.
    """
    args = [gatePair, args1, args2]
    kwargs = {}
    origin = rpcCall("IonGeneralMS", args, kwargs)
    return origin['result'], dictMatrixToNumpyMatrix(origin["unitary"], complex)



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Infidelity tools
"""

from numpy import all, ndarray, conjugate, transpose, shape, trace, dot, isclose, allclose
from scipy import linalg
from math import pow

from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.QOperator import dagger
from Quanlse.Utils.Functions import project


def unitaryInfidelity(uGoal: ndarray, uReal: ndarray, subSysNum: int) -> float:
    r"""
    Return the unitary infidelity of the ``uReal`` with ``uGoal``.
    The unitary infidelity is calculated using trace distance:

    :math:`g = 1 - \frac{1}{{\rm dim}(U)} \left| {\rm Tr} \left( U_{\rm goal}^{\dagger} U_{\rm real} \right) \right|.`

    :param uGoal: target unitary
    :param uReal: real unitary
    :param subSysNum: number of qubits
    :return: value of unitary infidelity
    """
    dimUGoal = shape(uGoal)[0]
    dimUReal = shape(uReal)[0]
    # Calculate qubit levels
    levelUGoal = int(pow(dimUGoal, 1 / subSysNum))
    levelUReal = int(pow(dimUReal, 1 / subSysNum))
    # We first check whether the dimension of the computational state is reasonable
    if levelUGoal == levelUReal:
        uGoal = transpose(conjugate(uGoal))
        return 1 - abs(trace(dot(uGoal, uReal))) / dimUGoal
    elif levelUGoal > levelUReal:
        uGoalProj = project(uGoal, subSysNum, levelUGoal, levelUReal)
        uGoalProj = transpose(conjugate(uGoalProj))
        return 1 - abs(trace(dot(uGoalProj, uReal))) / levelUReal
    else:
        uGoal = transpose(conjugate(uGoal))
        uRealProj = project(uReal, subSysNum, levelUReal, levelUGoal)
        return 1 - abs(trace(dot(uGoal, uRealProj))) / dimUGoal


def isRho(rho: ndarray, tol: float = 1e-7):
    r"""
    Check if the input matrix satisfies the following three requirements: Non-negative
    definite, Hermitian, Normalized.

    :param rho: the input matrix
    :param tol: the tolerance of the imprecision
    :return: -1 if not Hermitian; -2 if not normalized; -3 if not non-negative definite,
        1 when it satisfies the conditions of density operator.
    """

    # Check if rho is Hermitian
    if not allclose(rho - conjugate(transpose(rho)), 0.):
        return -1

    # Check if rho is normalized
    if not isclose(trace(rho), 1):
        return -2

    # Check if rho is non-negative definite
    if not all(linalg.eigvalsh(rho) > -tol):
        return -3

    return 1


def rhoInfidelity(rhoGoal: ndarray, rhoReal: ndarray) -> float:
    r"""
    Calculate the infidelity of the two density matrices.

   :param rhoGoal: the target final density matrix
   :param rhoReal: the real final density matrix
   :return: calculated infidelity of the input two density matrices
   """

    if rhoGoal.shape != rhoReal.shape:
        raise ArgumentError("The dimensions of two the input matrices are not matched.")

    rhoGoalSqrt = linalg.sqrtm(rhoGoal)

    return 1 - trace(linalg.sqrtm(rhoGoalSqrt @ rhoReal @ rhoGoalSqrt)) ** 2


def traceDistance(rhoGoal: ndarray, rhoReal: ndarray) -> float:
    r"""
    Calculate the trace distance between two density matrices.

    :param rhoGoal: the target final density matrix
    :param rhoReal: the real final density matrix
    :return: the trace distance of the two input two density matrices
    """

    if rhoGoal.shape != rhoReal.shape:
        raise ArgumentError("The dimensions of two the input matrices are not matched.")

    if not isRho(rhoGoal) or not isRho(rhoReal):
        raise ArgumentError("The input matrix doesn't meet the criteria of density matrix.")

    return 0.5 * trace(linalg.sqrtm(dagger(rhoGoal - rhoReal) @ (rhoGoal - rhoReal)))



#!/usr/bin/python3
# -*- coding: utf8 -*-
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Two-qubit gates Calibration
"""

import numpy
from math import pi
from numpy import array
from typing import Callable, Any, List, Optional

from Quanlse.QOperation.RotationGate import RX
from Quanlse.QOperation.FixedGate import CZ, H
from Quanlse.QPlatform.Error import ArgumentError
from Quanlse.Scheduler import Scheduler
from Quanlse.Simulator import PulseModel
from Quanlse.Utils.Functions import getPopulationOnQubit, basis, project


def caliSingleQubitGatesJob(sche: Scheduler, q0: int, q1: int, q0Para: Optional[List[float]],
                            q1Para: Optional[List[float]], returnMatrix: bool = False):
    """
    Calibrate the dynamical phase by tuning the Z pulse.

    :param sche: the Scheduler instance
    :param q0: the index of the first qubit to be calibrated
    :param q1: the index of the second qubit to be calibrated
    :param q0Para: the pulse parameters for the first qubit (piAmp and DRAG coefficient)
    :param q1Para: the pulse parameters for the second qubit (piAmp and DRAG coefficient)
    :param returnMatrix: returns the ideal matrix for the circuit when True
    :return: the QJob instance
    """

    if q0Para is not None:
        sche.conf["caliDataXY"][q0]["piAmp"] = q0Para[0]
        sche.conf["caliDataXY"][q0]["dragCoef"] = q0Para[1]

    if q1Para is not None:
        sche.conf["caliDataXY"][q1]["piAmp"] = q1Para[0]
        sche.conf["caliDataXY"][q1]["dragCoef"] = q1Para[1]

    matrixList = []
    jobList = sche.ham.createJobList()

    # Job 1: two H gates run simultaneously
    sche.clearCircuit()
    H(sche.Q[q0])
    H(sche.Q[q1])
    jobList.addJob(sche.schedule())
    if returnMatrix:
        matrixList.append(sche.getMatrix())

    # Job 2: two X/2 gates run simultaneously
    sche.clearCircuit()
    RX(pi / 2)(sche.Q[q0])
    RX(pi / 2)(sche.Q[q1])
    jobList.addJob(sche.schedule())
    if returnMatrix:
        matrixList.append(sche.getMatrix())

    # Job 3: two X gates run simultaneously
    sche.clearCircuit()
    RX(pi)(sche.Q[q0])
    RX(pi)(sche.Q[q1])
    jobList.addJob(sche.schedule())
    if returnMatrix:
        matrixList.append(sche.getMatrix())

    # Job 4: one X gate runs on the first qubit
    sche.clearCircuit()
    RX(pi)(sche.Q[q0])
    jobList.addJob(sche.schedule())
    if returnMatrix:
        matrixList.append(sche.getMatrix())

    return jobList, matrixList


def caliSingleQubitGates(sche: Scheduler, q0: int, q1: int, runner: Callable = None, bounds: List = None,
                         q0ParaInit: Optional[List[float]] = None, q1ParaInit: List[float] = None, options: Any = None):
    """
    Find the best qubit frequency.

    :param sche: the Scheduler instance
    :param q0: the index of the first qubit to be calibrated
    :param q1: the index of the second qubit to be calibrated
    :param runner: a callable function which can pass the QJob/QJobList instances and obtain the result
    :param bounds: the optimization bounds for each parameter
    :param q0ParaInit: specify a fixed frequency shift for qubit 0
    :param q1ParaInit: the list of the qubit 1 frequency shift (based on the present qubit frequency)
    :param options: options pass to the runner function
    """

    _, idealMatrixList = caliSingleQubitGatesJob(sche, q0, q1, None, None, True)

    def _singleQubitCaliLoss(x):
        """
        Obtain the calibration pulse jobs and calculate the loss function.
        """

        # Generate the pulse jobs for calibration
        caliJobs, _ = caliSingleQubitGatesJob(sche, q0, q1, [x[0], x[1]], [x[2], x[3]], False)

        # Simulate the calibration pulse
        if isinstance(sche, PulseModel):
            results = sche.simulate(jobList=caliJobs, options=options)
        else:
            if runner is None or not callable(runner):
                raise ArgumentError("Please input a callable runner!")
            results = runner(jobList=caliJobs, options=options)

        lossSum = 0.
        for _jobIdx, _idealMat in enumerate(idealMatrixList):
            # Obtain the output population value
            pop = project(numpy.square(abs(results[_jobIdx]["state"])).T[0], sche.subSysNum, sche.sysLevel, 2)

            # Obtain the ideal population
            stateIdeal = _idealMat @ basis(_idealMat.shape[0], 0).T[0]
            popIdeal = numpy.square(abs(stateIdeal))

            # Calculate the loss function
            lossSum += numpy.sum(numpy.abs(pop - popIdeal)) / len(pop)

        aveLoss = lossSum / len(idealMatrixList)
        return aveLoss

    from scipy.optimize import minimize
    InitParas = [q0ParaInit[0], q0ParaInit[1], q1ParaInit[0], q1ParaInit[1]]
    opt = minimize(_singleQubitCaliLoss, x0=numpy.array(InitParas), bounds=bounds, method="l-bfgs-b")
    return opt["x"][0], opt["x"][1], opt["x"][2], opt["x"][3], opt["fun"]


def czCaliCondPhaseJob(sche: Scheduler, q0: int, q1: int, q0ZAmp: float, q1ZAmp: float, czLen: float):
    """
    Generate the calibration pulses (QJobList instances) for the conditional phase by tuning the Z pulse.

    :param sche: the Scheduler instance
    :param q0: the index of the first qubit to be calibrated
    :param q1: the index of the second qubit to be calibrated
    :param q0ZAmp: the amplitude of Z pulse on the first qubit
    :param q1ZAmp: the amplitude of Z pulse on the second qubit
    :param czLen: the duration of the Z pulse
    :return: the QJobList instance
    """

    qubits = (q0, q1)

    # Set the pulse parameter
    sche.conf["caliDataCZ"][qubits]["q0ZAmp"] = q0ZAmp
    sche.conf["caliDataCZ"][qubits]["q1ZAmp"] = q1ZAmp
    sche.conf["caliDataCZ"][qubits]["czLen"] = czLen
    sche.conf["caliDataCZ"][qubits]["q0VZPhase"] = 0.
    sche.conf["caliDataCZ"][qubits]["q1VZPhase"] = 0.

    # -------------------------------------
    # 1. When qubit B is initialized as |1>
    # -------------------------------------
    sche.clearCircuit()
    # Add the X pulse
    RX(pi / 2)(sche.Q[q0])
    RX(pi)(sche.Q[q1])
    # Add the Z pulse
    CZ(sche.Q[q0], sche.Q[q1])
    # Add the X/2 pulse
    RX(pi / 2)(sche.Q[q0])
    # Generate the pulse
    job1 = sche.schedule()

    # -------------------------------------
    # 2. When qubit B is initialized as |0>
    # -------------------------------------
    sche.clearCircuit()
    # Add the X pulse
    RX(pi / 2)(sche.Q[q0])
    # Add the Z pulse
    CZ(sche.Q[q0], sche.Q[q1])
    # Add the X/2 pulse
    RX(pi / 2)(sche.Q[q0])
    # Generate the pulse
    job2 = sche.schedule()

    # Insert into the jobList
    jobList = sche.ham.createJobList()
    jobList.addJob(job1)
    jobList.addJob(job2)

    return jobList


def czCaliCondPhase(sche: Scheduler, q0: int, q1: int, runner: Callable = None, method: str = 'dual_annealing',
                    maxIter: int = 50, q0ZAmpInit: float = None, q1ZAmpInit: float = None, czLenInit: float = None,
                    options: Any = None):
    """
    Run the conditional phase calibration procedure for the controlled-Z gates.

    :param sche: the Scheduler instance
    :param q0: the index of the first qubit to be calibrated
    :param q1: the index of the second qubit to be calibrated
    :param runner: a callable function which can pass the QJob/QJobList instances and obtain the result
    :param method: the optimization method, can be chose from 'dual_annealing' or 'Nelder-Mead'
    :param maxIter: the maximum iteration
    :param q0ZAmpInit: the initial value of the amplitude of Z pulse on the first qubit
    :param q1ZAmpInit: the initial value of the amplitude of Z pulse on the second qubit
    :param czLenInit: the initial value of the duration of Z pulse
    :param options: options pass to the runner function
    """

    def _condPhaseCaliLoss(x):
        """
        Obtain the calibration pulse jobs and calculate the loss function.
        """

        # Generate the pulse jobs for calibration
        caliJobs = czCaliCondPhaseJob(sche, q0, q1, x[0], x[1], x[2])

        # Simulate the calibration pulse
        if isinstance(sche, PulseModel):
            results = sche.simulate(jobList=caliJobs, options=options)
        else:
            if runner is None or not callable(runner):
                raise ArgumentError("Please input a callable runner!")
            results = runner(jobList=caliJobs, options=options)

        # Obtain the output population value
        popB1 = numpy.square(abs(results[0]["state"])).T[0]
        popB0 = numpy.square(abs(results[1]["state"])).T[0]

        # Population of the first qubit in basis |0> when the second qubit is |1>
        pop0B1 = getPopulationOnQubit(popB1, 1, sche.sysLevel)

        # Population of the first qubit in basis |1> when the second qubit is |0>
        pop1B0 = getPopulationOnQubit(popB0, 1, sche.sysLevel)

        # Calculate the loss function
        loss = pop0B1[0] + pop1B0[1]
        return loss

    if method == 'dual_annealing':
        from scipy.optimize import dual_annealing
        opt = dual_annealing(_condPhaseCaliLoss, [(-5, 5), (-5, 5), (30, 50)], maxiter=maxIter)
    elif method == 'Nelder-Mead':
        from scipy.optimize import minimize
        if czLenInit is None or q0ZAmpInit is None or q1ZAmpInit is None:
            raise ArgumentError(f"You must specify initial values when using '{method}'!")
        opt = minimize(_condPhaseCaliLoss, array([q0ZAmpInit, q1ZAmpInit, czLenInit]), method=method)
    else:
        raise ArgumentError(f"Unsupported optimization method '{method}'!")

    return opt["x"][0], opt["x"][1], opt["x"][2], opt["fun"]


def czCaliDynamicalPhaseJob(sche: Scheduler, q0: int, q1: int, q0VZPhase: float, q1VZPhase: float):
    """
    Calibrate the dynamical phase by tuning the Z pulse.

    :param sche: the Scheduler instance
    :param q0: the index of the first qubit to be calibrated
    :param q1: the index of the second qubit to be calibrated
    :param q0VZPhase: the amplitude of the calibration phase on the first qubit
    :param q1VZPhase: the amplitude of the calibration phase on the second qubit
    :return: the QJob instance
    """

    qubits = (q0, q1)

    # Set the pulse parameter
    sche.conf["caliDataCZ"][qubits]["q0VZPhase"] = q0VZPhase
    sche.conf["caliDataCZ"][qubits]["q1VZPhase"] = q1VZPhase

    # -------------------------------------
    # 1. When qubit B is initialized as |1>
    # -------------------------------------
    sche.clearCircuit()
    # Add the H pulse
    H(sche.Q[q0])
    H(sche.Q[q1])
    # Add the Z pulse
    CZ(sche.Q[q0], sche.Q[q1])
    # Add the X/2 pulse
    H(sche.Q[q1])
    # Generate the pulse
    job = sche.schedule()

    return job


def czCaliDynamicalPhase(sche: Scheduler, q0: int, q1: int, runner: Callable = None, method: str = 'Nelder-Mead',
                         maxIter: int = 50, q0VZPhaseInit: float = None, q1VZPhaseInit: float = None,
                         options: Any = None):
    """
    Run the dynamical phase calibration procedure for the controlled-Z gates.

    :param sche: the Scheduler instance
    :param q0: the index of the first qubit to be calibrated
    :param q1: the index of the second qubit to be calibrated
    :param runner: a callable function which can pass the QJob/QJobList instances and obtain the result
    :param method: the optimization method, can be chose from 'dual_annealing' or 'Nelder-Mead'
    :param maxIter: the maximum iteration
    :param q0VZPhaseInit: the initial value of the amplitude of Z pulse on the first qubit
    :param q1VZPhaseInit: the initial value of the amplitude of Z pulse on the second qubit
    :param options: options pass to the runner function
    """

    def _dynamicalPhaseCaliLoss(x):
        """
        Obtain the calibration pulse jobs and calculate the loss function.
        """

        # Generate the pulse jobs for calibration
        caliJob = czCaliDynamicalPhaseJob(sche, q0, q1, x[0], x[1])

        # Simulate the calibration pulse
        if isinstance(sche, PulseModel):
            results = sche.simulate(job=caliJob, options=options)
        else:
            if runner is None or not callable(runner):
                raise ArgumentError("Please input a callable runner!")
            results = runner(job=caliJob, options=options)

        # Obtain the output population value
        pop = project(numpy.square(abs(results[0]["state"])).T[0], sche.subSysNum, sche.sysLevel, 2)

        # Obtain the ideal population
        uIdeal = sche.getMatrix()
        stateIdeal = uIdeal @ basis(uIdeal.shape[0], 0).T[0]
        popIdeal = numpy.square(abs(stateIdeal))

        # Calculate the loss function
        loss = numpy.sum(numpy.abs(pop - popIdeal)) / len(pop)
        return loss

    if method == 'dual_annealing':
        from scipy.optimize import dual_annealing
        opt = dual_annealing(_dynamicalPhaseCaliLoss, [(-2 * pi, 2 * pi), (-2 * pi, 2 * pi)], maxiter=maxIter)
    elif method == 'Nelder-Mead':
        from scipy.optimize import minimize
        if q0VZPhaseInit is None or q1VZPhaseInit is None:
            raise ArgumentError(f"You must specify 'q0VZPhaseInit' and 'q1VZPhaseInit' when using '{method}'!")
        opt = minimize(_dynamicalPhaseCaliLoss, array([q0VZPhaseInit, q1VZPhaseInit]), method=method)
    else:
        raise ArgumentError(f"Unsupported optimization method '{method}'!")

    return opt["x"][0], opt["x"][1], opt["fun"]



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Generally, when modeling a superconducting qubits system, we need to define the system's
Hamiltonian :math:`\hat{H}_{\rm sys}` firstly. The Hamiltonian, as the total energy, can
be used to describe the behavior of entire system:

:math:`\hat{H}_{\rm sys}(t) = \hat{H}_{\rm drift} + \hat{H}_{\rm coup} + \hat{H}_{\rm ctrl}(t)`

It typically contains three terms - the time-independent drift term describing the
individual qubits in the system, the coupling term describing the interaction between
qubits, and the time-dependent control term describing the control driving acting on qubits.

With the Hamiltonian constructed, we can simulate the evolution of the quantum system
by solving Schrödinger equation in the Heisenberg picture and then obtain the time-ordered
evolution operator :math:`U(t)`,

:math:`i\hbar\frac{{\rm \partial}U(t)}{{\rm \partial}t} = \hat{H}_{\rm sys}(t)U(t)`

A variety of functions and a complete set of pre-defined operators are provided in Quanlse,
which allows users to construct the Hamiltonian for large quantum systems with ease.

For more details, please visit https://quanlse.baidu.com/#/doc/tutorial-construct-ham.
"""

import copy
import pickle
import base64
from typing import Dict, Any, List, Union, Optional, Callable
from numpy import ndarray, shape, identity, size, kron, zeros, array, floor, transpose, conjugate
from numpy.linalg import eig

from Quanlse.QPlatform import Error
from Quanlse.QOperator import QOperator, create, destroy
from Quanlse.QWaveform import QWaveform, QJob, QJobList, QResult
from Quanlse.Utils.Functions import generateOperatorKey, tensor


class QHamiltonian:
    """
    Basic class of Hamiltonian, contains all information regarding a system.
    Users can create a QHamiltonian object to define the physical
    properties of a system. Constructing a Hamiltonian is essential to most Quanlse programs, we encourage the
    users to familiarize themselves with this step before advancing to the other features Quanlse provides.
    (see single-qubit gate tutorial for a complete example)

    :param subSysNum: size of the subsystem
    :param sysLevel: the energy levels of the system (support different levels for different qubits)
    :param dt: AWG sampling time
    :param title: a user-given title to the QHamiltonian object
    :param description: a user-given description
    """

    def __init__(self, subSysNum: int, sysLevel: Union[int, List[int]], dt: float,
                 title: str = "", description: str = "") -> None:
        """
        The constructor of the QHamiltonian class.
        """
        self.subSysNum = subSysNum  # type: int
        self.sysLevel = sysLevel  # type: int
        self.title = title  # type: str
        self.description = description  # type: str
        self.dt = dt  # type: float

        self._driftOperators = {}  # type: Dict[str, Union[QOperator, List[QOperator]]]
        self._collapseList = None

        # Create QJob instance
        self._waveJob = QJob(subSysNum, sysLevel, dt)  # type: QJob
        self._waveJob.parent = self

        self._collapseOperators = None  # type: Optional[List[ndarray]]
        self._driftCache = None  # type: Optional[ndarray]
        self._ctrlCache = {}  # type: Optional[Dict[str, Any]]
        self._dissipationSuperCache = None  # type: Optional[ndarray]

        # Save the waveforms with flag QWAVEFORM_FLAG_DO_NOT_CLEAR
        self._doNotClearFlagOperators = {}  # type: Dict[str, Any]
        self._doNotClearFlagWaves = {}  # type: Dict[str, Any]

        # When extracting subsystem,
        #     this property keeps tracks the relations
        #     between the original and the new indices.
        self.subSystemIndexMapping = None  # type: Optional[Dict[int, int]]

    def __str__(self) -> str:
        """
        Printer function for the QHamiltonian object.
        """
        returnStr = ""

        # Print basic information
        returnStr += f"(1) Basic Information of QHam `{'None' if self.title == '' else self.title}`:\n"
        returnStr += f"    - System Level: {self.sysLevel}\n"
        returnStr += f"    - Number of subsystems: {self.subSysNum}\n"
        returnStr += f"    - dt: {self.dt}\n"
        returnStr += f"    - Description: {'None' if self.description == '' else self.description}\n"

        # Print drift terms
        returnStr += f"(2) Drift (Time-independent) Terms:\n"
        if len(self._driftOperators.keys()) < 1:
            returnStr += f"    No drift terms.\n"
        else:
            for key in self._driftOperators.keys():
                if isinstance(self._driftOperators[key], list):
                    returnStr += f"    - `{key}` * {self._driftOperators[key][0].coef}\n"
                else:
                    returnStr += f"    - `{key}` * {self._driftOperators[key].coef}\n"

        # Print control terms
        returnStr += f"(3) Control (Time-dependent) Terms and Waveforms:\n"
        if len(self._waveJob.ctrlOperators.keys()) < 1:
            returnStr += f"    No control terms.\n"
        else:
            for key in self._waveJob.ctrlOperators.keys():
                returnStr += f"    - `{key}`: {len(self._waveJob.waves[key])} wave(s).\n"

        # Print Local oscillator setting
        returnStr += f"(4) Local Oscillator Settings:\n"
        if len(self._waveJob.LO.keys()) < 1:
            returnStr += f"    No LO setting.\n"
        else:
            for key in self._waveJob.LO.keys():
                _freq = self._waveJob.LO[key][0]
                _phase = self._waveJob.LO[key][1]
                returnStr += f"    - `{key}`: frequency is {_freq}, phase is {_phase}.\n"

        return returnStr

    @property
    def subSysNum(self) -> int:
        """
        Return the number of subsystems.
        """
        return self._subSysNum

    @subSysNum.setter
    def subSysNum(self, value: int):
        """
        Set the number of subsystems.
        """
        if not isinstance(value, int):
            raise Error.ArgumentError("subSysNum must be an integer!")
        if value < 1:
            raise Error.ArgumentError("subSysNum must be larger than 0!")
        self._subSysNum = value

    @property
    def sysLevel(self) -> Union[int, List[int]]:
        """
        Return the level of subsystems.
        """
        return self._sysLevel

    @sysLevel.setter
    def sysLevel(self, value: Union[int, List[int]]):
        """
        Set the level of subsystems.

        :param value: the number of subsystems - either an integer value or a list of integers.
        """
        if not isinstance(value, int) and not isinstance(value, list):
            raise Error.ArgumentError("sysLevel must be an integer or a list!")
        if isinstance(value, list) and min(value) < 2:
            raise Error.ArgumentError("All items in sysSize must be larger than 1!")
        if isinstance(value, int) and value < 2:
            raise Error.ArgumentError("All items in sysSize must be larger than 1!")
        self._sysLevel = value

    @property
    def ctrlCache(self) -> Dict[str, Any]:
        """
        Return the cache of the control terms.
        """
        return self._ctrlCache

    @ctrlCache.deleter
    def ctrlCache(self):
        """
        Delete the cache of the control terms.
        """
        del self._ctrlCache

    @property
    def driftCache(self) -> ndarray:
        """
        Return the cache of the drift terms.
        """
        return self._driftCache

    @driftCache.deleter
    def driftCache(self):
        """
        Delete the cache of the drift terms.
        """
        del self._driftCache

    @property
    def waveCache(self) -> Dict[str, Any]:
        """
        Return the cache of drift terms.
        """
        return self._waveJob.waveCache

    @waveCache.deleter
    def waveCache(self):
        del self._waveJob.waveCache

    @property
    def driftOperators(self) -> Dict[str, Union[QOperator, List[QOperator]]]:
        """
        Return the operators of the drift terms.
        """
        return self._driftOperators

    @driftOperators.deleter
    def driftOperators(self):
        """
        Delete the operators of the drift terms.
        """
        del self._driftOperators
        self._driftOperators = {}

    @property
    def doNotClearFlagOperators(self) -> Dict[str, Union[QOperator, List[QOperator]]]:
        """
        Return operators with flag QWAVEFORM_FLAG_DO_NOT_CLEAR.
        """
        return self._doNotClearFlagOperators

    @property
    def doNotClearFlagWaves(self) -> Dict[str, Union[QWaveform, List[QWaveform]]]:
        """
        Return the waves with flag QWAVEFORM_FLAG_DO_NOT_CLEAR.
        """
        return self._doNotClearFlagWaves

    @property
    def ctrlOperators(self) -> Dict[str, Union[QOperator, List[QOperator]]]:
        """
        Return operators of the control term.
        """
        return self._waveJob.ctrlOperators

    @property
    def collapseOperators(self) -> Optional[List[ndarray]]:
        """
        Return the list of the cache of the collapse operators for the Lindblad master equations.
        """
        return self._collapseOperators

    @collapseOperators.deleter
    def collapseOperators(self):
        """
        Delete the list of the cache of the collapse operators for the Lindblad master equations.
        """
        del self._collapseOperators
        self._collapseOperators = None

    @property
    def dissipationSuperCache(self):
        """
        Return the cache of dissipation super operator.
        """
        return self._dissipationSuperCache

    @dissipationSuperCache.deleter
    def dissipationSuperCache(self):
        """
        Delete the cache of the dissipation super operators.
        """
        del self._dissipationSuperCache

    @property
    def job(self) -> QJob:
        """
        Return a QJob object. The QJob object contains all the information regarding a quantum task.
        """
        return self._waveJob

    @job.setter
    def job(self, job: QJob):
        """
        Set a QJob object.

        :param job: a QJob object
        """
        # Check if the input waveList matches the system conf of the QHam.
        if job.sysLevel != self.sysLevel:
            raise Error.ArgumentError(f"sysLevel does not match ({job.sysLevel} and {self.sysLevel})!")
        if job.subSysNum != self.subSysNum:
            raise Error.ArgumentError(f"subSysNum does not match ({job.subSysNum} and {self.subSysNum})!")
        if not self.verifyWaveListObj(job):
            raise Error.ArgumentError("The input waveList is invalid for this ham, please check the"
                                      "system size or the dimension of the matrices.")
        # Save WaveList
        del self._waveJob
        self._waveJob = job
        self._waveJob.parent = self

    @job.deleter
    def job(self):
        """
        Delete a QJob object.
        """
        del self._waveJob

    def copy(self) -> "QHamiltonian":
        """
        Return a copy of current object.
        """
        return copy.deepcopy(self)

    def createJobList(self) -> QJobList:
        """
        Return an instance of a QJobList of the same system properties.
        """
        newJobList = QJobList(subSysNum=self.subSysNum, sysLevel=self.sysLevel, dt=self.dt)
        newJobList.LO = self.job.LO
        return newJobList

    def createJob(self) -> QJob:
        """
        Return an instance of a QJob object of the same system properties.
        """
        newJob = QJob(subSysNum=self.subSysNum, sysLevel=self.sysLevel, dt=self.dt)
        newJob.ctrlOperators = copy.deepcopy(self.job.ctrlOperators)
        newJob.LO = self.job.LO
        return newJob

    def addDrift(self, operator: Union[QOperator, Callable, List[QOperator], List[Callable]],
                 onSubSys: Union[int, List[int]], coef: float = None, name: str = None) -> None:
        r"""
        Add drift terms to the Hamiltonian.

        **Example 1** (add drift terms):

        Given a three-qubit system, the users could run the following code:

        .. code-block:: python

                ham.addDrift(operator=[number(2), number(2)], onSubSys=[0, 2], coef=omega1, name="Operator1")

        The code above allows the users to efficiently add the specified drift operator to the Hamiltonian:
        :math:`H = \omega_1 (\hat{a}^{\dagger}\hat{a}) \otimes \hat{I} \otimes (\hat{a}^{\dagger}\hat{a}),`

        where :math:`\hat{a},\hat{a}^{\dagger}` are the annihilation and creation operators; :math:`\hat{I}`
        is the identity operator.

        :param operator: the respective operator(s)
        :param onSubSys: qubit indexes that the term acts upon
        :param coef: the corresponding coefficient of the term
        :param name: a user-given name to the term for identifying purposes
        :return: None
        """

        _operators = []
        if isinstance(operator, list):
            for opIdx, op in enumerate(operator):
                if isinstance(op, QOperator):
                    # Input operator is an QOperator instance
                    _operators.append(copy.deepcopy(op))
                else:
                    # Input operator is callable
                    if isinstance(self.sysLevel, list):
                        _operators.append(op(self.sysLevel[onSubSys[opIdx]]))
                    else:
                        _operators.append(op(self.sysLevel))
        else:
            if isinstance(operator, QOperator):
                # Input operator is an QOperator instance
                _operators = copy.deepcopy(operator)
            else:
                # Input operator is callable
                if isinstance(self.sysLevel, list):
                    _operators = operator(self.sysLevel[onSubSys])
                else:
                    _operators = operator(self.sysLevel)

        if onSubSys is None:
            # Input the complete matrices directly.
            if isinstance(self._sysLevel, int):
                dim = self._sysLevel ** self._subSysNum
            elif isinstance(self._sysLevel, list):
                dim = 1
                for lvl in self._sysLevel:
                    dim = dim * lvl
            else:
                raise Error.ArgumentError(f"Wrong type of self._sysLevel ({type(self._sysLevel)}).")

            if not shape(_operators.matrix) == (dim, dim):
                raise Error.ArgumentError(f"Dimension of operator ({dim}, {dim}) does not match the sysLevel.")

            operatorForSave = _operators
        else:
            # Verify if operator and onSubSys have the same size
            if (isinstance(_operators, list) and not isinstance(onSubSys, list)) or \
                    (not isinstance(_operators, list) and isinstance(onSubSys, list)):
                raise Error.ArgumentError("The type of operator should be the same as onSubSys's!")

            if isinstance(_operators, list) and isinstance(onSubSys, list):
                if len(_operators) != len(onSubSys):
                    raise Error.ArgumentError(f"The size of operator ({len(_operators)}) != that "
                                              f"of onSubSys ({len(onSubSys)})!")

            # Verify the range and set the onSubSys
            if isinstance(onSubSys, int):
                if onSubSys >= self._subSysNum:
                    raise Error.ArgumentError(f"onSubSys ({onSubSys}) is larger than the "
                                              f"subSysNum {self._subSysNum}.")
                _operators.onSubSys = onSubSys
                operatorForSave = _operators
            elif isinstance(onSubSys, list):
                if max(onSubSys) >= self._subSysNum:
                    raise Error.ArgumentError(f"onSubSys ({onSubSys}) is larger than the "
                                              f"subSysNum {self._subSysNum}.")
                # Sort the operators according to on onSubSys
                sortedIndex = list(array(onSubSys).argsort())
                sortedOperators = [_operators[i] for i in sortedIndex]
                sortedOnSubSys = [onSubSys[i] for i in sortedIndex]
                # Set onSubSys
                for i in range(len(sortedIndex)):
                    sortedOperators[i].onSubSys = sortedOnSubSys[i]
                operatorForSave = sortedOperators
            else:
                raise Error.ArgumentError("onSubSys should be an integer or a list!")

        # Set the name if given
        if name is not None:
            opKey = name
        else:
            opKey = generateOperatorKey(self.subSysNum, operatorForSave)

        # Verify the uniqueness of the combination of the name and onSubSys
        if opKey in self._driftOperators.keys():
            raise Error.ArgumentError(f"Name ({opKey}) exists!")

        if coef is not None:
            if not isinstance(coef, float):
                raise Error.ArgumentError(f"coef should be float instead of {type(coef)}")
            if isinstance(_operators, list):
                for op in _operators:
                    op.coef = coef
            else:
                _operators.coef = coef

        self._driftOperators[opKey] = operatorForSave

    def addCoupling(self, onSubSys: List, g: float) -> None:
        r"""
        Add a coupling term to the Hamiltonian.

        :param onSubSys: the index list that the term works upon
        :param g: coupling strength
        :return: None.

        Two qubit indexes should be provided; and the coupling term is added between the two:

        :math:`H_{\rm coupling} = g (\hat{a}_{\rm q1}^{\dagger} \hat{a}_{\rm q2} +
        \hat{a}_{\rm q1} \hat{a}_{\rm q2}^{\dagger}),`

        where :math:`\hat{a}_q,\hat{a}_q^{\dagger}` are the annihilation and creation
        operators on the :math:`q`-th qubit.

        **Example 1** (add coupling terms):

        Given a three-qubit system, the users could run the following code:

        .. code-block:: python

                ham.addCoupling(onSubSys=[0, 2], g=0.0227 * 2 * pi)

        The code above allows the users to efficiently add the specified drift operator to the Hamiltonian:

        :math:`H = 0.14262830631 (\hat{a}_0 \otimes \hat{I} \otimes \hat{a}_2^{\dagger} + \hat{a}_0^{\dagger}
        \otimes \hat{I} \otimes \hat{a}_2).`
        """
        if len(onSubSys) != 2:
            raise Error.ArgumentError("Coupling term should be defined on subsystems of two qubits!")

        if isinstance(self._sysLevel, int):
            d = self._sysLevel
            # Construct matrices
            matrices = [destroy(d), create(d)]
            matricesHc = [create(d), destroy(d)]
        else:
            matrices = [destroy(self._sysLevel[onSubSys[0]]), create(self._sysLevel[onSubSys[1]])]
            matricesHc = [create(self._sysLevel[onSubSys[0]]), destroy(self._sysLevel[onSubSys[1]])]
        # Add drift terms
        self.addDrift(matrices, onSubSys=onSubSys, coef=g)
        self.addDrift(matricesHc, onSubSys=onSubSys, coef=g)

    def addWave(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                onSubSys: Union[int, List[int]] = None, waves: Union[QWaveform, List[QWaveform]] = None,
                t0: Union[int, float] = None, strength: Union[int, float] = 1.0,
                freq: Optional[Union[int, float]] = None, phase: Optional[Union[int, float]] = None,
                phase0: Optional[Union[int, float]] = None, name: str = None, tag: str = None, flag: int = None) \
            -> None:
        r"""
        This method adds a control term to the Hamiltonian with a specified waveform.

        :param operators: wave operator
        :param onSubSys: what subsystem the wave is acting upon
        :param waves: a QWaveform object
        :param t0: start time
        :param strength: wave strength
        :param freq: wave frequency shift
        :param phase: pulse phase shift (will accumulate during the entire pulse execution)
        :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
        :param name: a user-given name to the wave
        :param tag: wave tag indicating purpose
        :param flag: wave flag
        :return: None
        """
        self._waveJob.addWave(operators, onSubSys, waves, t0, strength, freq, phase, phase0, name, tag, flag)

    def appendWave(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                   onSubSys: Union[int, List[int]] = None, waves: Union[QWaveform, List[QWaveform]] = None,
                   strength: Union[int, float] = 1.0, freq: Optional[Union[int, float]] = None,
                   phase: Optional[Union[int, float]] = None, phase0: Optional[Union[int, float]] = None,
                   name: str = None, tag: str = None, shift: float = 0.0, compact: bool = True):
        """
        This method appends control terms and waveforms to a QJob object. Unlike `addWave()`, this function will append
        the waveform in the end of the existing waveforms, hence ignore the `t0` parameter of rhe `QWaveform` object.

        :param operators: wave operator
        :param onSubSys: what subsystem the wave is acting upon
        :param waves: a QWaveform object
        :param strength: wave strength
        :param freq: wave frequency shift
        :param phase: pulse phase shift (will accumulate during the entire pulse execution)
        :param phase0: pulse phase shift (will not accumulate during the entire pulse execution)
        :param name: a user-given name to the wave
        :param tag: wave tag indicating purpose
        :param shift: time shift of the added wave
        :param compact: the added wave will be left aligned in specified control terms if True.
        :return: None
        """
        self._waveJob.appendWave(operators, onSubSys, waves, strength, freq, phase, phase0,
                                 name, tag, shift, compact)

    def clearWaves(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]] = None,
                   onSubSys: Union[int, List[int]] = None, names: Union[str, List[str]] = None,
                   tag: str = None) -> None:
        """
        This method removes all waveforms in the specified control terms.
        If names is None, this method will remove waveforms in all control terms.

        :param operators: the corresponding operator(s)
        :param onSubSys: qubit indexes that the term acts upon
        :param names: the user-given name for identifying purposes
        :param tag: a user-given tag marking the pulse's usage
        :return: None.
        """
        self.job.clearWaves(operators, onSubSys, names, tag)

    def setLO(self, operators: Union[QOperator, Callable, List[QOperator], List[Callable]],
              onSubSys: Union[int, List[int]], name: str = None, freq: float = 1.0, phase: float = 0.0):
        """
        Add local oscillator for specific control terms.

        :param operators: QOperator object(s)
        :param onSubSys: what subsystem the wave is acting upon
        :param name: user-defined operator name
        :param freq: the frequency of the local oscillator
        :param phase: the phase between the local oscillator and the waveform
        """
        self._waveJob.setLO(operators, onSubSys, name, freq, phase)

    def plot(self, names: Union[str, List[str]] = None, tag: str = None, xUnit: str = 'ns', yUnit: str = 'Amp (a.u.)',
             color: Union[str, List[str]] = None, dark: bool = False) -> None:
        """
        Plot waves.

        :param names: "names" to plot
        :param tag: a tag marking the pulse's usage
        :param xUnit: x axis unit to be displayed
        :param yUnit: y axis unit to be displayed
        :param color: a list of colors for each pulse
            (from mint, blue, red, green, yellow, black, pink, cyan, purple, darkred, orange, brown, pink and teal)
            If pulse consists more than 250 cuts, switch to slice
        :param dark: enables a dark-themed mode
        :return: None.
        """
        self.job.plot(names, tag, xUnit, yUnit, color, dark)

    def setCollapse(self, cList: Optional[Dict[int, List[ndarray]]]) -> None:
        """
        Define the list of collapse operators for the Lindblad equation.

        :param cList: a list of collapse operators
        :return: None.
        """

        for index, operators in cList.items():
            if type(operators) is not list:
                raise Error.ArgumentError(f"the value of the key in cList should be a list type object")
            for operator in operators:
                if isinstance(self.sysLevel, int):
                    if shape(operator)[0] != self.sysLevel:
                        raise Error.ArgumentError("the dimension of the collapse operator doesn't equal"
                                                  " to the dimension of the state!")
                if isinstance(self.sysLevel, list):
                    if shape(operator)[0] != self.sysLevel[index]:
                        raise Error.ArgumentError("the dimension of the collapse operator doesn't equal"
                                                  " the dimension of the state!")
        self._collapseList = cList

    def verifyWaveListObj(self, waveJob: QJob) -> bool:
        """
        Check whether the QWaveList object is valid for this QHam object.

        :param waveJob: QWaveList to be checked
        :return: true of false indicating its validity
        """
        for key in waveJob.ctrlOperators.keys():
            if isinstance(waveJob.ctrlOperators[key], list):
                ops = waveJob.ctrlOperators[key]
            elif isinstance(waveJob.ctrlOperators[key], QOperator):
                ops = [waveJob.ctrlOperators[key]]
            else:
                raise Error.ArgumentError("The input QWaveList object is invalid!")
            # Check every operator
            for op in ops:
                if op.onSubSys >= self._subSysNum:
                    return False
                if isinstance(self._sysLevel, int):
                    if max(op.matrix.shape) > self._sysLevel:
                        return False
                else:
                    if max(op.matrix.shape) > self._sysLevel[op.onSubSys]:
                        return False
        return True

    def buildCache(self) -> None:
        """
        Build Cache for waveList and operators.

        :return: None
        """
        self.clearCache()
        self.buildOperatorCache()
        self._waveJob.buildWaveCache()
        if self._collapseList is not None:
            self.buildCollapseCache()
            self.buildDissipationSuperCache()

    def buildOperatorCache(self) -> None:
        r"""
        Save the drift/coupling/control terms in a more efficient way for further usage.
        Note that this function will recursively process all the terms added to drift/coupling/control terms and
        save them to cache.

        :return: None
        """
        sysLevel = self.sysLevel
        subSysNum = self.subSysNum

        # Generator the operator for all of the drift terms
        driftMatList = []
        for key in self._driftOperators.keys():
            if isinstance(self._driftOperators[key], list):
                mat = self._driftOperators[key][0].coef * self._generateOperator(self._driftOperators[key])
                driftMatList.append(mat)
            else:
                mat = self._driftOperators[key].coef * self._generateOperator(self._driftOperators[key])
                driftMatList.append(mat)
        # Sum all the drift terms and save to the cache.
        if isinstance(sysLevel, int):
            driftTotal = zeros((sysLevel ** subSysNum, sysLevel ** subSysNum), dtype=complex)
        else:
            dim = 1
            for i in sysLevel:
                dim = dim * i
            driftTotal = zeros((dim, dim), dtype=complex)
        for mat in driftMatList:
            driftTotal = driftTotal + mat
        self._driftCache = driftTotal

        # Generator the pulse sequences for all of the control terms.
        for key in self._waveJob.ctrlOperators.keys():
            ctrlsOp = self._waveJob.ctrlOperators[key]
            self._ctrlCache[key] = self._generateOperator(ctrlsOp)
        for key in self.doNotClearFlagOperators:
            if key not in self._ctrlCache:
                ctrlsOp = self.doNotClearFlagOperators[key]
                self._ctrlCache[key] = self._generateOperator(ctrlsOp)

    def buildCollapseCache(self) -> None:
        r"""
        Generate and save the cache of the collapse operators.

        :return: None
        """

        levelList = []

        if isinstance(self.sysLevel, list):
            levelList = self.sysLevel

        elif isinstance(self.sysLevel, int):
            levelList = [self.sysLevel for _ in range(self.subSysNum)]

        cList = self._collapseList

        cacheList = []

        for index, operators in cList.items():
            matList = [identity(level) for level in levelList]

            for operator in operators:
                matList[index] = operator
                mat = tensor(matList)
                cacheList.append(mat)

        self._collapseOperators = cacheList

    def buildDissipationSuperCache(self) -> None:
        """
        Generate the dissipation super operator in the Liouville form.
        """
        cList = self._collapseOperators

        dim = None

        # Calculate the dimension of the Hilbert space
        if isinstance(self.sysLevel, int):
            dim = self.sysLevel ** self.subSysNum

        elif isinstance(self.sysLevel, list):
            dim = 1
            for i in self.sysLevel:
                dim = dim * i

        idn = identity(dim, dtype=complex)

        ld = zeros((dim * dim, dim * dim), dtype=complex)

        if cList is not None:
            cdagList = [transpose(conjugate(c)) for c in cList]
            for c, cdag in zip(cList, cdagList):
                ld += kron(c, c) - 0.5 * kron(cdag @ c, idn) - 0.5 * kron(idn, cdag @ c)

        self._dissipationSuperCache = ld

    def _generateOperator(self, operator: Union[QOperator, List[QOperator]]) -> ndarray:
        """
        Generate the operator of the system in the complete Hilbert space by taking tensor products.

        :param operator: a list of operator(s)
        :return: the operator after taking the tensor products of the input operator
        """
        sysLevel = self.sysLevel
        subSysNum = self.subSysNum
        # Generate matrices
        if isinstance(operator, list):
            matrices = []
            for op in operator:
                matrices.append(op.matrix)
        else:
            matrices = operator.matrix
        # Each subsystem of the system has the same energy level.
        if isinstance(sysLevel, int):
            # We first define the identity matrix to fill un-assigned subsystems
            idMat = identity(sysLevel, dtype=complex)
            if isinstance(operator, QOperator):
                if size(matrices) == (sysLevel, sysLevel):
                    raise Error.ArgumentError("Dimension of matrix does not match the system Level.")
                # The operator is on only one subsystem.
                if operator.onSubSys == 0:
                    # This operator is on the first subsystem.
                    finalOperator = matrices
                    for i in range(1, subSysNum):
                        finalOperator = kron(finalOperator, idMat)
                else:
                    # This operator is not on the first subsystem.
                    finalOperator = idMat
                    for i in range(1, operator.onSubSys):
                        finalOperator = kron(finalOperator, idMat)
                    finalOperator = kron(finalOperator, matrices)
                    for i in range(operator.onSubSys + 1, subSysNum):
                        finalOperator = kron(finalOperator, idMat)
                return finalOperator
            elif isinstance(operator, list):
                finalOperator = []
                onSubSys = []
                for op in operator:
                    onSubSys.append(op.onSubSys)
                for i in range(subSysNum):
                    if i == 0:
                        # On the first subsystem
                        if i in onSubSys:
                            matrixIndex = onSubSys.index(i)
                            finalOperator = matrices[matrixIndex]
                            operatorSize = shape(matrices[matrixIndex])
                            if not (operatorSize == (sysLevel, sysLevel)):
                                raise Error.ArgumentError(f"Dim of input matrix {operatorSize} does not match"
                                                          f" with the system level ({sysLevel}).")
                        else:
                            finalOperator = idMat
                    else:
                        # Not on the first subsystem
                        if i in onSubSys:
                            matrixIndex = onSubSys.index(i)
                            operatorSize = shape(matrices[matrixIndex])
                            if not (operatorSize == (sysLevel, sysLevel)):
                                raise Error.ArgumentError(f"Dim of input matrix {operatorSize} does not match"
                                                          f" with the system level ({sysLevel}).")
                            finalOperator = kron(finalOperator, matrices[matrixIndex])
                        else:
                            finalOperator = kron(finalOperator, idMat)
                return finalOperator

            else:
                raise Error.ArgumentError("Variable onSubSys should be a list or an int value.")
        # The sysLevel is a list of different energy levels for multiple subsystems
        elif isinstance(sysLevel, list):
            # Create a list of identities of different dimension for each subsystem of different energy level
            idMat = [identity(i, dtype=complex) for i in sysLevel]
            # The operator is acting on only one subsystem.
            if isinstance(operator, QOperator):
                if not (matrices.shape == (sysLevel[operator.onSubSys], sysLevel[operator.onSubSys])):
                    raise Error.ArgumentError("Dimension of matrix does not match the system Level.")
                # The operator is acting on the first subsystem.
                if operator.onSubSys == 0:
                    finalOperator = matrices
                    for i in range(1, subSysNum):
                        finalOperator = kron(finalOperator, idMat[i])
                else:
                    # This operator is not acting on the first subsystem.
                    finalOperator = idMat[0]
                    for i in range(1, operator.onSubSys):
                        finalOperator = kron(finalOperator, idMat[i])
                    finalOperator = kron(finalOperator, matrices)
                    for i in range(operator.onSubSys + 1, subSysNum):
                        finalOperator = kron(finalOperator, idMat[i])
                return finalOperator
            # The operator is acting on multiple subsystems.
            elif isinstance(operator, list):
                finalOperator = []
                onSubSys = []
                for op in operator:
                    onSubSys.append(op.onSubSys)
                for i in range(subSysNum):
                    if i == 0:
                        # Acting on the first subsystem
                        if i in onSubSys:
                            matrixIndex = onSubSys.index(i)
                            finalOperator = matrices[matrixIndex]
                            operatorSize = shape(matrices[matrixIndex])
                            if not (operatorSize == (sysLevel[i], sysLevel[i])):
                                raise Error.ArgumentError(f"Dim of input matrix {operatorSize} does not match"
                                                          f" with the system level ({sysLevel[i]}).")
                        else:
                            finalOperator = idMat[i]
                    else:
                        # Not acting on the first subsystem
                        if i in onSubSys:
                            matrixIndex = onSubSys.index(i)
                            operatorSize = shape(matrices[matrixIndex])
                            if not (operatorSize == (sysLevel[i], sysLevel[i])):
                                raise Error.ArgumentError(f"Dim of input matrix {operatorSize} does not match"
                                                          f" with the system level ({sysLevel[i]}).")
                            finalOperator = kron(finalOperator, matrices[matrixIndex])
                        else:
                            finalOperator = kron(finalOperator, idMat[i])
                return finalOperator

            else:
                raise Error.ArgumentError("Variable onSubSys should be a list or an int value.")

    def clearCache(self):
        """
        Clear cache.
        """
        self._ctrlCache = {}
        self._driftCache = None
        self._collapseOperators = None
        self._waveJob.clearCache()

    def subSystem(self, onSubSys: Union[int, List[int]]) -> 'QHamiltonian':
        """
        This method extracts a subsystem from the target Hamiltonian.

        :param onSubSys: a list of qubit indexes
        :return: subsystem's QHamiltonian object
        """
        subHam = copy.deepcopy(self)
        subHam.clearCache()
        # Information about the sub system
        if isinstance(onSubSys, int):
            subQubits = 1
            indexMapping = {onSubSys: 0}
            if onSubSys >= self._subSysNum:
                raise Error.ArgumentError("onSubSys should be less than the subSysNum of this QHamiltonian.")
        elif isinstance(onSubSys, list):
            subQubits = len(onSubSys)
            # qubit index mapping
            indexMapping = {}
            for newId, originalId in enumerate(onSubSys):
                indexMapping[originalId] = newId
                if originalId >= self._subSysNum:
                    raise Error.ArgumentError("all onSubSys items should be less than the subSysNum "
                                              "of this QHamiltonian.")
        else:
            raise Error.ArgumentError("Unsupported input of qubitNum, it should be an int value or a list.")

        # Update the sysLevel property
        if isinstance(self._sysLevel, list):
            if isinstance(onSubSys, int):
                subHam.sysLevel = self._sysLevel[onSubSys]
                subHam.job.sysLevel = self._sysLevel[onSubSys]
            else:
                subHam.sysLevel = [self._sysLevel[qIndex] for qIndex in onSubSys]
                subHam.job.sysLevel = [self._sysLevel[qIndex] for qIndex in onSubSys]

        # Set the mapping relationship
        subHam.subSystemIndexMapping = copy.deepcopy(indexMapping)

        def mapping(index: Union[int, List[int]]) -> Union[int, List[int]]:
            """
            Map the `onQubits` index(es) from the original Ham to the extracted Ham.

            :param index: 'onQubits' index(es) from the original Ham
            :return: mapped index(es) onto the extracted Ham
            """
            if isinstance(index, int):
                return indexMapping[index]
            else:
                mappedList = []
                for item in index:
                    mappedList.append(indexMapping[item])
                return mappedList

        def allIn(listA: Union[int, List[int]], listB: Union[int, List[int]]) -> bool:
            """
            Check whether all the items in listB are in listA.

            :param listA: listA
            :param listB: listB
            :return: a bool value
            """
            if isinstance(listA, int):
                listA = [listA]
            if isinstance(listB, int):
                return listB in listA
            else:
                for item in listB:
                    if item not in listA:
                        return False
                return True

        # Trim the drift terms
        del subHam.driftOperators
        for key in self._driftOperators:
            drifts = self._driftOperators[key]  # type: Union[QOperator, List[QOperator]]
            if isinstance(drifts, list):
                _coef = drifts[0].coef
                originalOnSubSys = [op.onSubSys for op in drifts]
            else:
                _coef = drifts.coef
                originalOnSubSys = copy.deepcopy(drifts.onSubSys)
            if allIn(onSubSys, originalOnSubSys):
                subHam.addDrift(copy.deepcopy(drifts), mapping(originalOnSubSys), coef=_coef)

        # Trim the control terms
        subHam.job.clearWaves()
        for key in self.job.ctrlOperators:
            ctrls = self.job.ctrlOperators[key]  # type: Union[QOperator, List[QOperator]]
            if isinstance(ctrls, list):
                originalOnSubSys = [op.onSubSys for op in ctrls]
            else:
                originalOnSubSys = ctrls.onSubSys
            if allIn(onSubSys, originalOnSubSys):
                wavesDict = self.job.searchWave(ctrls, originalOnSubSys)
                for waveKey in wavesDict.keys():
                    # Update the waveform setup
                    subHam.job.addWave(ctrls, mapping(originalOnSubSys), waves=wavesDict[waveKey])

        # Trim the DoNotClear terms
        subHam.doNotClearFlagWaves.clear()
        subHam.doNotClearFlagOperators.clear()

        for key in self.doNotClearFlagWaves:
            doNotClrCtrl = self.doNotClearFlagOperators[key]  # type: Union[QOperator, List[QOperator]]
            if isinstance(doNotClrCtrl, list):
                originalOnSubSys = [op.onSubSys for op in doNotClrCtrl]
            else:
                originalOnSubSys = doNotClrCtrl.onSubSys
            if allIn(onSubSys, originalOnSubSys):
                wavesDict = QJob.searchWaveTool(self.doNotClearFlagWaves, self.subSysNum, self.sysLevel,
                                                doNotClrCtrl, originalOnSubSys)
                for waveKey in wavesDict.keys():
                    # Update the waveform setup
                    subHam.addWave(doNotClrCtrl, mapping(originalOnSubSys), waves=wavesDict[waveKey])

        # Trim the collapse terms
        cList = copy.deepcopy(self._collapseList)
        subCList = None
        if cList is not None:
            subCList = {}
            for originalId, newId in indexMapping.items():
                if originalId not in cList.keys():
                    pass
                else:
                    subCList[newId] = cList[originalId]

        # Update the basic information
        subHam._collapseList = subCList
        subHam.subSysNum = subQubits
        subHam.job.subSysNum = subQubits

        return subHam

    def eigen(self, t: Optional[ndarray] = None):
        """
        Calculate the eigenvalues and eigenvectors of the given Hamiltonian.

        :param t: The time at which the Hamiltonian's eigenvectors and eigenvalues are to compute.
        :return: If t is none, it returns (n, n) Hamiltonian's eigenvalues of shape (n, ) in the
        ascending order and the corresponding eigenvectors matrix of shape (n, n); If t is an (m, ) array,
        it returns list of (m, n) eigenvalues and (m, n, n) eigenvectors.
        """

        # Build cache of the Hamiltonian
        self.buildCache()
        drift = self.driftCache
        ctrl = self.ctrlCache
        wave = self.job.waveCache
        dt = self.dt
        maxDt = self.job.endTimeDt

        def _computeEigen(matrix):
            # Solve eigenvalues problem
            eigenVals, eigenVecs = eig(matrix)
            # Rearrange eigenvalues and eigenvectors in the ascending order
            sortedIndex = eigenVals.argsort()
            eigenVecs = eigenVecs[:, sortedIndex]
            eigenVals = eigenVals[sortedIndex]
            return eigenVals.real, eigenVecs

        if t is None:
            # Solve eigenvalues problem
            vals, vecs = _computeEigen(drift)
        else:
            valsList = []
            vecsList = []
            if len(t) > maxDt:
                raise Error.ArgumentError('The length of time exceeds the number of maximum steps')
            indexDt = floor(t / dt)
            indexDt = indexDt.astype(int)
            for nowDt in indexDt:
                if nowDt >= maxDt:
                    nowDt = nowDt - 1
                ham = drift
                for key in ctrl:
                    ham += ctrl[key] * wave[key][nowDt]
                evals, evecs = _computeEigen(ham)
                valsList.append(evals.tolist())
                vecsList.append(evecs.tolist())
            vals = array(valsList)
            vecs = array(vecsList)

        # Clear cache
        self.clearCache()
        return vals, vecs

    def subSystemIndicesInverse(self) -> Dict[int, int]:
        """
        In self.subSystem, users can extract a subsystem of the original system,
        thus, the qubit indices will change. For example, for a system with indices [0, 1, 2, 3],
        we extract the subsystem constituted by subsystems 1 and 3, then the index of
        subsystem 1 changes to 0; 3 changes to 1; the mapping in self.subSystemIndexMapping is:
        {1: 0, 3: 1}.
        We can use this function to obtain the inverse mapping, the inverse mapping of the mapping
        above is:
        {0: 1, 1: 3}

        :return: a dictionary containing the 'inverse mapping'
        """
        inverseMapping = {}
        for key in self.subSystemIndexMapping:
            inverseMapping[self.subSystemIndexMapping[key]] = key
        return inverseMapping

    def outputInverseJob(self, subSysNum: int, sysLevel: int = None, dt: float = None) -> QJob:
        """
        Return an inverse-mapped Job. (see subSystemIndicesInverse())

        :param subSysNum: subsystem's size
        :param sysLevel: subsystem's energy level
        :param dt: time interval
        :return: returned QJob object
        """
        # Get the job
        inverseMapping = self.subSystemIndicesInverse()
        _sysLevel = self.sysLevel if sysLevel is None else sysLevel
        _dt = self.dt if dt is None else dt
        job = QJob(subSysNum, _sysLevel, _dt)
        for opKey in self.job.ctrlOperators.keys():
            operators = self.job.ctrlOperators[opKey]
            onSubSysInSubHam = self.job.ctrlOperators[opKey].onSubSys
            if isinstance(operators, list):
                onSubSysInOriginalHam = []
                for index in onSubSysInSubHam:
                    onSubSysInOriginalHam.append(inverseMapping[index])
            elif isinstance(operators, QOperator):
                onSubSysInOriginalHam = inverseMapping[onSubSysInSubHam]
            else:
                raise Error.ArgumentError("Unsupported type of operators.")
            # Add the waves into the new QJob instance.
            for wave in self.job.waves[opKey]:
                job.addWave(operators, onSubSysInOriginalHam, wave)
        return job

    def simulate(self, job: QJob = None, state0: ndarray = None, recordEvolution: bool = False, shot: int = None,
                 isOpen: bool = False, jobList: QJobList = None, refreshCache: bool = True, accelerate: bool = False,
                 adaptive: bool = False, tolerance: float = 0.01) -> QResult:
        """
        Calculate the unitary evolution operator with a given Hamiltonian. This function supports
        both single-job and batch-job processing.

        To activate acceleration, please install ``Numba`` JIT compiler, please visit its official website
        https://numba.pydata.org/ for more details.

        :param job: the QJob object to simulate.
        :param state0: the initial state vector. If None is given, this function will return the time-ordered
                       evolution operator, otherwise returns the final state vector.
        :param recordEvolution: the detailed procedure will be recorded if True
        :param shot: return the population of the eigenstates when ``shot`` is provided
        :param isOpen: simulate the evolution using Lindblad equation
        :param jobList: a job list containing the waveform list
        :param refreshCache: it will neither clear nor rebuild cache if refreshCache is set to be True
        :param adaptive: use the adaptive solver if True
        :param tolerance: the greatest error of approximation for the adaptive solver
        :param accelerate: use numba expm if true

        :return: result dictionary (or a list of result dictionaries when ``jobList`` is provided)

        This function does provide the option of simulating on local devices. However, Quanlse also provides
        cloud computing services which are significantly faster.

        **Example 1** (single-job processing):

        If ``jobList`` is None, the waveforms in ham will be used for simulation, and return a result dictionary.

        .. code-block:: python

                result = ham.simulate()

        **Example 2** (batch-job processing):

        If ``jobList`` is not None, simulation will use the waveform list for batch-job processing provided in
        jobList, and return a list of result dictionaries.

        .. code-block:: python

                result = ham.simulate(ham, jobList=jobs)

        Here, jobs is a QJobList object, indicating the the waveforms to be used.
        """

        resultList = QResult("built-in simulator")

        if jobList is None:
            jobList = self.createJobList()
            if job is None:
                jobList.addJob(self.job)
            else:
                jobList.addJob(job)

        # Transverse all the jobs
        _tmpHam = copy.deepcopy(self)
        for waveList in jobList.jobs:
            # Clear all waves
            _tmpHam.job.clearWaves()
            _tmpHam.job.LO = waveList.LO
            # Add waves in current job
            _tmpHam.job.appendJob(waveList)
            # Build cache
            if refreshCache:
                _tmpHam.buildCache()
            # Start calculation
            if isOpen:
                if not adaptive:
                    from Quanlse.Utils.ODESolver import solverOpenSystem
                    result = solverOpenSystem(_tmpHam, state0=state0, recordEvolution=recordEvolution,
                                              accelerate=accelerate)
                else:
                    raise Error.ArgumentError("Adaptive solver is not available when solving open system.")
            else:
                if not adaptive:
                    from Quanlse.Utils.ODESolver import solverNormal
                    result = solverNormal(_tmpHam, state0=state0, shot=shot, recordEvolution=recordEvolution,
                                          accelerate=accelerate)
                else:
                    from Quanlse.Utils.ODESolver import solverAdaptive
                    result = solverAdaptive(_tmpHam, state0=state0, shot=shot, tolerance=tolerance,
                                            accelerate=accelerate)
            resultList.append(result)
            if refreshCache:
                self.clearCache()
        return resultList

    def doNotClearWaveFunctionsToSequences(self, maxEndTime: float = None) -> Dict[str, List[QWaveform]]:
        """
        Convert all functions into sequence (including the QWaveform objects) that is serializable.

        :param maxEndTime: maximum ending time
        :return: dictionary containing the converted functions
        """
        newList = {}
        for waveKey in self._doNotClearFlagWaves:
            # Add QWaveform
            newList[waveKey] = []
            for wave in self._doNotClearFlagWaves[waveKey]:
                seqWave = wave.waveFunctionToSequence(self.dt, maxEndTime=maxEndTime)
                newList[waveKey].append(seqWave)
        return newList

    def dump(self, maxEndTime: float = None) -> str:
        """
        Return a base64 encoded string.

        :param maxEndTime: maximum ending time
        :return: encoded string
        """
        obj = copy.deepcopy(self)
        obj._waveJob = None

        if self._waveJob is not None:
            del obj._waveJob
            obj._waveJob = self._waveJob.waveFunctionsToSequences(maxEndTime=maxEndTime)

        if self._doNotClearFlagWaves is not None:
            obj._doNotClearFlagWaves = None
            obj._doNotClearFlagWaves = self.doNotClearWaveFunctionsToSequences(maxEndTime=maxEndTime)
            for opKey in obj._doNotClearFlagWaves.keys():
                for wave in obj._doNotClearFlagWaves[opKey]:
                    if not hasattr(wave, "omega"):
                        wave.omega = wave.freq
                        wave.phi = 0.
                    if not hasattr(wave, "phi"):
                        wave.phi = 0. if wave.phase0 is None else wave.phase0

        if obj.job is not None:
            obj.job.parent = None

        # For compatibility with previous versions
        for opKey in obj.job.waves.keys():
            for wave in obj.job.waves[opKey]:
                if not hasattr(wave, "omega"):
                    wave.omega = wave.freq
                    wave.phi = 0.
                if not hasattr(wave, "phi"):
                    wave.phi = 0. if wave.phase0 is None else wave.phase0

        byteStr = pickle.dumps(obj)
        base64str = base64.b64encode(byteStr)
        del obj
        return base64str.decode()

    def clone(self) -> 'QHamiltonian':
        """
        Return the copy of the object
        """
        return copy.deepcopy(self)

    @staticmethod
    def load(base64Str: str) -> 'QHamiltonian':
        """
        Create object from base64 encoded string.

        :param base64Str: a base64 encoded string
        :return: QHam object
        """
        byteStr = base64.b64decode(base64Str.encode())
        obj = pickle.loads(byteStr)  # type: QHamiltonian

        if obj.job is not None:
            # Set job's parent
            obj.job.parent = obj
            # For compatibility with previous versions
            for opKey in obj.job.waves.keys():
                for wave in obj.job.waves[opKey]:
                    if not hasattr(wave, "freq"):
                        wave.freq = wave.omega if hasattr(wave, "omega") else None
                    if not hasattr(wave, "phase0"):
                        wave.phase0 = wave.phi if hasattr(wave, "phi") else None
                    if not hasattr(wave, "phase"):
                        wave.phase = None
        return obj



#!/usr/bin/python3
# -*- coding: utf8 -*-

# Copyright (c) 2020 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Readout simulator Template: 3-qubit readout simulator.
"""

from Quanlse.Simulator import ReadoutModel
from Quanlse.Simulator.PulseSim3Q import pulseSim3Q
from typing import Union
from math import pi


def readoutSim3Q(dt: Union[int, float] = 20.) -> ReadoutModel:
    r"""
    Return a template of 3-qubit readout model.

    :param dt: the sampling time period.

    :return: a readout simulator ReadoutModel object.
    """

    dt = dt  # The sampling time
    kappa = 0.0020 * (2 * pi)  # the system decay rate to the environment
    gain = 1000000.
    pulseModel = pulseSim3Q(dt=0.2)

    # resonator level
    resonatorLevel = 3  # resonator level
    resonatorFreq = {
        0: 7.104 * (2 * pi),
        1: 7.014 * (2 * pi),
        2: 7.214 * (2 * pi)
    }

    # qubit-resonator coupling strength
    coupling = {
        0: 0.134 * (2 * pi),
        1: 0.112 * (2 * pi),
        2: 0.121 * (2 * pi)
    }

    model = ReadoutModel(pulseModel=pulseModel, resonatorFreq=resonatorFreq, level=resonatorLevel,
                         coupling=coupling, dissipation=kappa, gain=gain, dt=dt)

    return model




#!/usr/bin/python3
# -*- coding: utf8 -*-
# Copyright (c) 2021 Baidu, Inc. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script implements various extrapolation methods such as the Richardson extrapolation and linear extrapolation.
"""

from scipy import optimize
import numpy as np
from collections import Iterable

from Quanlse.QPlatform import Error


def extrapolate(rescalingCoes: Iterable,
                expectations: Iterable,
                type: str = 'richardson',
                order: int = None,
                a0: float = None,
                sysFunc: bool = True) -> float:
    r"""
    Return the zero-noise extrapolation result (when the rescaling coefficient is zero) using different
    extrapolation strategies.


    :math:`\left\{\lambda_j, E_j \right\}_{1}^m \rightarrow\lim_{\lambda \to 0} E(\lambda)`

    :param rescalingCoes: Iterable, shape (m,).
        a series of rescaling coefficients (:math:`\lambda`)
    :param expectations: Iterable, shape (m,).
        a series of expectations corresponding to 'rescalingCoes'
    :param type: str, default: 'linear'.
        strategy for extrapolating, options: ['linear', 'polynomial', 'richardson', 'poly-exponential', 'exponential']
        or its simplified notations: ['linear', 'poly', 'richard', 'poly-exp', 'exp']
    :param order: int, default: None.
        default is None means that when using `poly-exp`-type or `poly`-type strategy,
        result of a specific order should specify its value
    :param a0: float, default: None.
        default is None means that when using 'poly-exp'-type strategy and the first term is unknown
    :param sysFunc: bool, default: True.
        True implies using the default np.polyfit method, otherwise, use the implemented linearRegression method
    :return: value of expectation when the rescaling coefficient is zero

    References
    ----------
    .. [1] Giurgica-Tiron, T., et al. (2020). Digital zero noise extrapolation for quantum error mitigation: 306-316.
    .. [2] Temme, K., et al. (2017). "Error mitigation for short-depth quantum circuits."
            Physical Review Letters 119(18): 180509.
    .. [3] Wikipedia, "Polynomial regression",
           https://en.wikipedia.org/wiki/Polynomial_regression
    .. [4] Wikipedia, "Polynomial interpolation",
           https://en.wikipedia.org/wiki/Polynomial_interpolation
    """
    typeOptional = ['linear', 'polynomial', 'richardson', 'poly-exponential', 'exponential']
    typeOptionalSimple = ['linear', 'poly', 'richard', 'poly-exp', 'exp']
    if len(rescalingCoes) != len(expectations):
        raise Error.ArgumentError('The length of "rescalingCoes" and "expectations" should be equal!')
    if (type not in typeOptional) and (type not in typeOptionalSimple):
        raise Error.ArgumentError('Please specify the "type" argument from {}'.format(typeOptional))
    # variables definition
    lams = np.array(rescalingCoes)
    Es = np.array(expectations)

    # number of samples
    m = len(lams)
    # extrapolating procedure
    if type == 'linear':
        if m == 2:
            # simply two-point extrapolating
            slope = (Es[1] - Es[0]) / (lams[1] - lams[0])
            res = Es[0] + slope * (0 - lams[0])
        else:
            # linear fitting
            if sysFunc:
                res = np.polyfit(lams, Es, 1)[-1]
            else:
                res = linearRegression(lams, Es)[0]
    elif type == 'richardson' or type == 'richard':
        # (m-1)th-order Richardson Extrapolation
        # 1) solve linear equations
        A = np.array([lams ** i for i in range(m)])
        b = np.zeros_like(lams)
        b[0] = 1
        # 2) compute extrapolated result
        coeffs = np.linalg.solve(A, b)
        res = sum(coeffs * Es)
    elif type == 'polynomial' or type == 'poly':
        if order is not None:
            # return extrapolation result of the specific order
            if order >= m - 1:
                raise Error.ArgumentError(
                    'Please limit value of "order" lower than "m-1" (m is length of "expectations")')
            # polynomial basis function transformation
            if sysFunc:
                res = np.polyfit(lams, Es, order)[-1]
            else:
                res = polyRegression(lams, Es, order)[0]
        else:
            # return extrapolation result of polynomial
            # order: [2, 3, ..., m-2] (list including m-3 elements)
            orders = range(2, m - 1)
            if sysFunc:
                res = [np.polyfit(lams, Es, d)[-1] for d in orders]
            else:
                res = [polyRegression(lams, Es, d)[0] for d in orders]
    elif type == 'poly-exponential' or type == 'poly-exp':
        if a0 is not None:  # a0 is already known
            sign = 1 if np.alltrue(Es - a0 > 0) else -1
            EsLog = np.log(np.abs(Es - a0))
            if order is not None:
                if order >= m - 1:
                    raise Error.ArgumentError('Please limit value of "order" '
                                              'lower than "m-1" (m is length of "expectations")')
                # polynomial basis function transformation
                if sysFunc:
                    z0 = np.polyfit(lams, EsLog, order)[-1]
                else:
                    z0 = polyRegression(lams, EsLog, order)[0]
                res = a0 + sign * np.exp(z0)
            else:
                # return extrapolation result of poly-exponential
                # order: [2, 3, ..., m-2] (list including m-3 elements)
                orders = range(2, m - 1)
                res = []
                for d in orders:
                    if sysFunc:
                        z0 = np.polyfit(lams, EsLog, d)[-1]
                    else:
                        z0 = polyRegression(lams, EsLog, d)[0]
                    res.append(a0 + sign * np.exp(z0))
        else:
            raise Error.ArgumentError("The {} extrapolation method with a0 "
                                      "is unknown is under construction.".format(type))
    elif type == 'exponential' or type == 'exp':
        if a0 is not None:  # a0 is already known
            sign = 1 if np.alltrue(Es - a0 > 0) else -1
            EsLog = np.log(np.abs(Es - a0))
            if m == 2:
                # simply two-point extrapolating
                slope = (EsLog[1] - EsLog[0]) / (lams[1] - lams[0])
                z0 = slope * (0 - lams[0]) + EsLog[0]
                res = a0 + sign * np.exp(z0)
            else:
                if sysFunc:
                    z0 = np.polyfit(lams, EsLog, 1)[-1]
                else:
                    z0 = linearRegression(lams, EsLog)[0]
                res = a0 + sign * np.exp(z0)
        else:
            # numeric fitting using scipy.optimize
            paras, _ = optimize.curve_fit(expSingle, lams, Es)
            res = paras[0] + paras[1]

    return res


def expSingle(x, a, b, c):
    r"""
    Function prototype used for optimize.curve_fit, with the expression :math:`a + b e^{- c x}`.
    """
    return a + b * np.exp(- c * x)


def linearRegression(x, y):
    r"""
    Linear regression method. Fit the given data points to the following linear model:

    :math:`y = \beta_0 + \beta_1     x`

    :param x: a list of independent variables
    :param y: a list of dependent variables
    :return: An array of coefficients: :math:`\beta_0` (intercept) and :math:`\beta_1` (slope)
    """
    x = np.array(x).ravel()
    y = np.array(y).ravel()
    m = len(x)
    xAvg = np.mean(x)
    yAvg = np.mean(y)
    Sxx = np.sum(x ** 2) - m * xAvg ** 2
    Sxy = np.sum(x * y) - m * xAvg * yAvg
    return np.array([yAvg - xAvg * Sxy / Sxx, Sxy / Sxx])


def polyRegression(x: list, y: list, d: int):
    r"""
    Polynomial regression method to the d-th order. Fit the given data points to the following linear model:

    :math:`y = \sum_{i=0}^d a_i x^i`

    :param x: a list of independent variables
    :param y: a list of dependent variables
    :param d: the polynomial regression order
    :return: an array of polynomial regression coefficients :math:`[a_0, a_1, \cdots, a_d]`
    """
    x = np.array(x).ravel()
    y = np.array(y).ravel()
    X = []  # shape: [m, d+1]
    for k in range(d + 1):
        X.append(np.expand_dims(x, axis=1))
    X = np.concatenate(X, axis=1)
    return np.dot(np.dot(np.linalg.pinv(np.dot(X.T, X)), X.T), y)
